#
# Copyright (c) 2010, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This module contains an abstraction of a MySQL server object used
by multiple utilities. It also contains helper methods for common
server operations used in multiple utilities.
"""

import os
import re
import socket
import string
import subprocess
import tempfile
import threading

import mysql.connector

from mysql.connector.errorcode import CR_SERVER_LOST
from mysql.utilities.exception import (ConnectionValuesError, UtilError,
                                       UtilDBError, UtilRplError)
from mysql.utilities.common.user import User
from mysql.utilities.common.tools import (delete_directory, execute_script,
                                          ping_host)
from mysql.utilities.common.ip_parser import (parse_connection, hostname_is_ip,
                                              clean_IPv6, format_IPv6)


_FOREIGN_KEY_SET = "SET foreign_key_checks = {0}"
_AUTOCOMMIT_SET = "SET AUTOCOMMIT = {0}"
_GTID_ERROR = ("The server %s:%s does not comply to the latest GTID "
               "feature support. Errors:")


def get_connection_dictionary(conn_info):
    """Get the connection dictionary.

    The method accepts one of the following types for conn_info:

        - dictionary containing connection information including:
          (user, passwd, host, port, socket)
        - connection string in the form: user:pass@host:port:socket or
                                         login-path:port:socket
        - an instance of the Server class

    conn_info[in]          Connection information

    Returns dict - dictionary for connection (user, passwd, host, port, socket)
    """
    if conn_info is None:
        return conn_info
    conn_val = {}
    if isinstance(conn_info, dict) and 'host' in conn_info:
        conn_val = conn_info
    elif isinstance(conn_info, Server):
        # get server's dictionary
        conn_val = conn_info.get_connection_values()
    elif isinstance(conn_info, basestring):
        # parse the string
        conn_val = parse_connection(conn_info, None)
    else:
        raise ConnectionValuesError("Cannot determine connection information"
                                    " type.")

    return conn_val


def _print_connection(prefix, conn_info):
    """Print connection information

    The method accepts one of the following types for conn_info:

        - dictionary containing connection information including:
          (user, passwd, host, port, socket)
        - connection string in the form: user:pass@host:port:socket or
                                         login-path:port:socket
        - an instance of the Server class

    conn_info[in]          Connection information
    """
    conn_val = get_connection_dictionary(conn_info)
    print "# %s on %s: ..." % (prefix, conn_val["host"]),


def get_local_servers(all_proc=False, start=3306, end=3333,
                      datadir_prefix=None):
    """Check to see if there are any servers running on the local host.

    This method attempts to locate all running servers. If provided, it will
    also limit the search to specific ports of datadirectory prefixes.

    This method uses ps for posix systems and netstat for Windows machines
    to determine the list of running servers.

    For posix, it matches on the datadir and if datadir is the path for the
    test directory, the server will be added to the list.

    For nt, it matches on the port in the range starting_port,
    starting_port + 10.

    all_proc[in]        If True, find all processes else only user processes
    start[in]           For Windows/NT systems: Starting port value to search.
                        Default = 3306
    end[in]             For Windows/NT systems: Ending port value to search.
                        Default = 3333
    datadir_prefix[in]  For posix systems, if not None, find only those servers
                        whose datadir starts with this prefix.

    Returns list - tuples of the form: (process_id, [datadir|port])
    """
    processes = []
    if os.name == "posix":
        tmp_file = tempfile.TemporaryFile()
        if all_proc:
            subprocess.call(["ps", "-A"], stdout=tmp_file)
        else:
            subprocess.call(["ps", "-f"], stdout=tmp_file)
        tmp_file.seek(0)
        for line in tmp_file.readlines():
            mysqld_safe = False
            mysqld = False
            datadir = False
            grep = False
            datadir_arg = ""
            proginfo = string.split(line)
            for arg in proginfo:
                if "datadir" in arg:
                    datadir = True
                    datadir_arg = arg
                if "mysqld" in arg:
                    mysqld = True
                if "mysqld_safe" in arg:
                    mysqld_safe = True
                if "grep" in arg:
                    grep = True
            # Check to see if this is a mysqld server and not mysqld_safe proc
            if ((mysqld and datadir) or (mysqld and not grep)) and \
               not mysqld_safe:
                # If provided, check datadir prefix
                if all_proc:
                    proc_id = proginfo[0]
                else:
                    proc_id = proginfo[1]
                if datadir_prefix is not None:
                    if datadir_prefix in datadir_arg:
                        processes.append((proc_id, datadir_arg[10:]))
                else:
                    processes.append((proc_id, datadir_arg[10:]))
    elif os.name == "nt":
        f_out = open("portlist", 'w+')
        execute_script("netstat -anop tcp", "portlist")
        f_out = open("portlist", 'r')
        for line in f_out.readlines():
            proginfo = string.split(line)
            if proginfo:
                # Look for port on either local or foreign address
                port = proginfo[1][proginfo[1].find(":") + 1:]
                if proginfo[1][0] == '0' and port.isdigit():
                    if int(port) >= start and int(port) <= end:
                        processes.append((proginfo[4], port))
                        break
                if len(proginfo) > 2:
                    port = proginfo[2][proginfo[2].find(":") + 1:]
                    if port.isdigit() and \
                       int(port) >= int(start) and int(port) <= int(end):
                        processes.append((proginfo[4], port))
                        break
        f_out.close()
        os.unlink("portlist")
    return processes


def get_server(name, values, quiet):
    """Connect to a server and return Server instance

    If the name is 'master' or 'slave', the connection will be made via the
    Master or Slave class else a normal Server class shall be used.

    name[in]           name of the server
    values[in]         dictionary of connection values
    quiet[in]          if True, do not print messages

    Returns Server class instance
    """
    from mysql.utilities.common.replication import Master, Slave

    server_conn = None

    # Try to connect to the MySQL database server.
    if not quiet:
        _print_connection(name, values)

    server_options = {
        'conn_info': values,
        'role': name,
    }
    if name.lower() == 'master':
        server_conn = Master(server_options)
    elif name.lower() == 'slave':
        server_conn = Slave(server_options)
    else:
        server_conn = Server(server_options)
    server_conn.connect()

    return server_conn


def _require_version(server, version):
    """Check version of server

    server[in]         Server instance
    version[in]        minimal version of the server required

    Returns boolean - True = version Ok, False = version < required
    """
    if version is not None and server is not None:
        major, minor, rel = version.split(".")
        if not server.check_version_compat(major, minor, rel):
            return False
    return True


def get_server_state(server, host, pingtime=3, verbose=False):
    """Return the state of the server.

    This method returns one of the following states based on the
    criteria shown.

      UP   - server is connected
      WARN - server is not connected but can be pinged
      DOWN - server cannot be pinged nor is connected

    server[in]     Server class instance
    host[in]       host name to ping if server is not connected
    pingtime[in]   timeout in seconds for ping operation
                   Default = 3 seconds
    verbose[in]    if True, show ping status messages
                   Default = False

    Returns string - state
    """
    if verbose:
        print "# Attempting to contact %s ..." % host,
    if server is not None and server.is_alive():
        if verbose:
            print "Success"
        return "UP"
    elif ping_host(host, pingtime):
        if verbose:
            print "Server is reachable"
        return "WARN"
    if verbose:
        print "FAIL"
    return "DOWN"


def connect_servers(src_val, dest_val, options=None):
    """Connect to a source and destination server.

    This method takes two groups of --server=user:password@host:port:socket
    values and attempts to connect one as a source connection and the other
    as the destination connection. If the source and destination are the
    same server and the unique parameter is False, destination is set to None.

    The method accepts one of the following types for the src_val and dest_val:

        - dictionary containing connection information including:
          (user, passwd, host, port, socket)
        - connection string in the form: user:pass@host:port:socket or
                                         login-path:port:socket
        - an instance of the Server class

    src_val[in]        source connection information
    dest_val[in]       destination connection information
    options[in]        options to control behavior:
        quiet          do not print any information during the operation
                       (default is False)
        version        if specified (default is None), perform version
                       checking and fail if server version is < version
                       specified - an exception is raised
        src_name       name to use for source server
                       (default is "Source")
        dest_name      name to use for destination server
                       (default is "Destination")
        unique         if True, servers must be different when dest_val is
                       not None (default is False)

    Returns tuple (source, destination) where
            source = connection to source server
            destination = connection to destination server (set to None)
                          if source and destination are same server
            if error, returns (None, None)
    """
    if options is None:
        options = {}
    quiet = options.get("quiet", False)
    src_name = options.get("src_name", "Source")
    dest_name = options.get("dest_name", "Destination")
    version = options.get("version", None)
    charset = options.get("charset", None)

    source = None
    destination = None

    # Get connection dictionaries
    src_dict = get_connection_dictionary(src_val)
    if "]" in src_dict['host']:
        src_dict['host'] = clean_IPv6(src_dict['host'])
    dest_dict = get_connection_dictionary(dest_val)
    if dest_dict and "]" in dest_dict['host']:
        dest_dict['host'] = clean_IPv6(dest_dict['host'])

    # Add character set
    if src_dict and charset:
        src_dict["charset"] = charset
    if dest_dict and charset:
        dest_dict["charset"] = charset

    # Check for uniqueness - dictionary
    if options.get("unique", False) and dest_dict is not None:
        dupes = False
        if "unix_socket" in src_dict and "unix_socket" in dest_dict:
            dupes = (src_dict["unix_socket"] == dest_dict["unix_socket"])
        else:
            dupes = (src_dict["port"] == dest_dict["port"]) and \
                    (src_dict["host"] == dest_dict["host"])
        if dupes:
            raise UtilError("You must specify two different servers "
                            "for the operation.")

    # If we're cloning so use same server for faster copy
    cloning = dest_dict is None or (src_dict == dest_dict)

    # Connect to the source server and check version
    if isinstance(src_val, Server):
        source = src_val
    else:
        source = get_server(src_name, src_dict, quiet)
        if not quiet:
            print "connected."
    if not _require_version(source, version):
        raise UtilError("The %s version is incompatible. Utility "
                        "requires version %s or higher." %
                        (src_name, version))

    # If not cloning, connect to the destination server and check version
    if not cloning:
        if isinstance(dest_val, Server):
            destination = dest_val
        else:
            destination = get_server(dest_name, dest_dict, quiet)
            if not quiet:
                print "connected."
        if not _require_version(destination, version):
            raise UtilError("The %s version is incompatible. Utility "
                            "requires version %s or higher." %
                            (dest_name, version))
    elif not quiet and dest_dict is not None and \
            not isinstance(dest_val, Server):
        _print_connection(dest_name, src_dict)
        print "connected."
    return (source, destination)


def test_connect(conn_info, throw_errors=False):
    """Test connection to a server.

    The method accepts one of the following types for conn_info:

        - dictionary containing connection information including:
          (user, passwd, host, port, socket)
        - connection string in the form: user:pass@host:port:socket or
                                         login-path:port:socket
        - an instance of the Server class

    conn_info[in]          Connection information

    throw_errors           throw any errors found during the test,
                           false by default.

    Returns True if connection success, False if error
    """
    # Parse source connection values
    try:
        src_val = get_connection_dictionary(conn_info)
    except Exception as err:
        raise ConnectionValuesError("Server connection values invalid: {0}."
                                    "".format(err))
    try:
        conn_options = {
            'quiet': True,
            'src_name': "test",
            'dest_name': None,
        }
        s = connect_servers(src_val, None, conn_options)
        s[0].disconnect()
    except UtilError:
        if throw_errors:
            raise
        return False
    return True


def check_hostname_alias(server1_vals, server2_vals):
    """Check to see if the servers are the same machine by host name.

    server1_vals[in]   connection dictionary for server1
    server2_vals[in]   connection dictionary for server2

    Returns bool - true = server1 and server2 are the same host
    """
    server1 = Server({'conn_info': server1_vals})
    server2 = Server({'conn_info': server2_vals})

    return (server1.is_alias(server2.host) and
            int(server1.port) == int(server2.port))


def stop_running_server(server, wait=10, drop=True):
    """Stop a running server.

    This method will stop a server using the mysqladmin utility to
    shutdown the server. It also destroys the datadir.

    server[in]          Server instance to clone
    wait[in]            Number of wait cycles for shutdown
                        default = 10
    drop[in]            If True, drop datadir

    Returns - True = server shutdown, False - unknown state or error
    """
    # Nothing to do if server is None
    if server is None:
        return True

    # Build the shutdown command
    cmd = ""
    res = server.show_server_variable("basedir")
    mysqladmin_client = "mysqladmin"
    if not os.name == "posix":
        mysqladmin_client = "mysqladmin.exe"
    mysqladmin_path = os.path.normpath(os.path.join(res[0][1], "bin",
                                                    mysqladmin_client))
    if not os.path.exists(mysqladmin_path):
        mysqladmin_path = os.path.normpath(os.path.join(res[0][1], "client",
                                                        mysqladmin_client))
    if not os.path.exists(mysqladmin_path) and not os.name == 'posix':
        mysqladmin_path = os.path.normpath(os.path.join(res[0][1],
                                                        "client/debug",
                                                        mysqladmin_client))
    if not os.path.exists(mysqladmin_path) and not os.name == 'posix':
        mysqladmin_path = os.path.normpath(os.path.join(res[0][1],
                                                        "client/release",
                                                        mysqladmin_client))
    cmd += mysqladmin_path
    if server.socket is None and server.host == 'localhost':
        server.host = '127.0.0.1'
    cmd += " shutdown --user=%s --host=%s " % (server.user, server.host)
    if server.passwd:
        cmd += "--password=%s " % server.passwd
    # Use of server socket only works with 'localhost' (not with 127.0.0.1).
    if server.socket and server.host == 'localhost':
        cmd += "--socket=%s " % server.socket
    else:
        cmd += "--port=%s " % server.port
    res = server.show_server_variable("datadir")
    datadir = res[0][1]

    # Kill all connections so shutdown will work correctly
    res = server.exec_query("SHOW PROCESSLIST")
    for row in res:
        if not row[7] or not row[7].upper().startswith("SHOW PROCESS"):
            try:
                server.exec_query("KILL CONNECTION %s" % row[0])
            except UtilDBError:  # Ok to ignore KILL failures
                pass

    # disconnect user
    server.disconnect()

    # Stop the server
    f_null = os.devnull
    f_out = open(f_null, 'w')
    proc = subprocess.Popen(cmd, shell=True,
                            stdout=f_out, stderr=f_out)
    ret_val = proc.wait()
    f_out.close()

    # if shutdown doesn't work, exit.
    if int(ret_val) != 0:
        return False

    # If datadir exists, delete it
    if drop:
        delete_directory(datadir)

    if os.path.exists("cmd.txt"):
        try:
            os.unlink("cmd.txt")
        except:
            pass

    return True


class Server(object):
    """The Server class can be used to connect to a running MySQL server.
    The following utilities are provided:

        - Connect to the server
        - Retrieve a server variable
        - Execute a query
        - Return list of all databases
        - Return a list of specific objects for a database
        - Return list of a specific objects for a database
        - Return list of all indexes for a table
        - Read SQL statements from a file and execute
    """

    def __init__(self, options=None):
        """Constructor

        The method accepts one of the following types for options['conn_info']:

            - dictionary containing connection information including:
              (user, passwd, host, port, socket)
            - connection string in the form: user:pass@host:port:socket or
                                             login-path:port:socket
            - an instance of the Server class

        options[in]        options for controlling behavior:
            conn_info      a dictionary containing connection information
                           (user, passwd, host, port, socket)
            role           Name or role of server (e.g., server, master)
            verbose        print extra data during operations (optional)
                           default value = False
            charset        Default character set for the connection.
                           (default None)
        """
        if options is None:
            options = {}

        assert not options.get("conn_info") is None

        self.verbose = options.get("verbose", False)
        self.db_conn = None
        self.host = None
        self.role = options.get("role", "Server")
        conn_values = get_connection_dictionary(options.get("conn_info"))
        try:
            self.host = conn_values["host"]
            self.user = conn_values["user"]
            self.passwd = conn_values["passwd"] \
                if "passwd" in conn_values else None
            self.socket = conn_values["unix_socket"] \
                if "unix_socket" in conn_values else None
            self.port = 3306
            if conn_values["port"] is not None:
                self.port = int(conn_values["port"])
            self.charset = options.get("charset",
                                       conn_values.get("charset", None))
        except KeyError:
            raise UtilError("Dictionary format not recognized.")
        self.connect_error = None
        # Set to TRUE when foreign key checks are ON. Check with
        # foreign_key_checks_enabled.
        self.fkeys = None
        self.autocommit = None
        self.read_only = False
        self.aliases = []
        self.grants_enabled = None
        self._version = None

    @classmethod
    def fromServer(cls, server):
        """ Create a new server instance from an existing one

        Factory method that will allow the creation of a new server instance
        from an existing server.

        server[in] - instance object that must be instance of the Server class
                     or a subclass.

        Returns an instance of the calling class as a result.
        """

        if isinstance(server, Server):
            options = {"conn_info": server.get_connection_values(),
                       "role": server.role,
                       "verbose": server.verbose,
                       "charset": server.charset}
            return cls(options)
        else:
            raise TypeError("The server argument's type is neither Server nor "
                            "a subclass of Server")

    def is_alive(self):
        """Determine if connection to server is still alive.

        Returns bool - True = alive, False = error or cannot connect.
        """
        res = True
        try:
            if self.db_conn is None:
                res = False
            else:
                # ping and is_connected only work partially, try exec_query
                # to make sure connection is really alive
                retval = self.db_conn.is_connected()
                if retval:
                    self.exec_query("SHOW DATABASES")
                else:
                    res = False
        except:
            res = False
        return res

    def is_alias(self, host_or_ip):
        """Determine if host_or_ip is an alias for this host

        host_or_ip[in] host or IP number to check

        Returns bool - True = host_or_ip is an alias
        """
        def get_aliases(host):
            """Gets the aliases for the given host
            """
            aliases = [clean_IPv6(host)]
            if hostname_is_ip(clean_IPv6(host)):  # IP address
                try:
                    my_host = socket.gethostbyaddr(clean_IPv6(host))
                    aliases.append(my_host[0])
                    # socket.gethostbyname_ex() does not work with ipv6
                    if (not my_host[0].count(":") < 1 or
                       not my_host[0] == "ip6-localhost"):
                        host_ip = socket.gethostbyname_ex(my_host[0])
                    else:
                        addrinfo = socket.getaddrinfo(my_host[0], None)
                        host_ip = ([socket.gethostbyaddr(addrinfo[0][4][0])],
                                   [fiveple[4][0] for fiveple in addrinfo],
                                   [addrinfo[0][4][0]])
                except (socket.gaierror, socket.herror,
                        socket.error) as err:
                    host_ip = ([], [], [])
                    if self.verbose:
                        print("WARNING: IP lookup by address failed for {0},"
                              "reason: {1}".format(host, err.strerror))
            else:
                try:
                    # server may not really exist.
                    host_ip = socket.gethostbyname_ex(host)
                except (socket.gaierror, socket.herror,
                        socket.error) as err:
                    if self.verbose:
                        print("WARNING: hostname: {0} may not be reachable, "
                              "reason: {1}".format(host, err.strerror))
                    return aliases
                extend_aliases(aliases, [host_ip[0]])
                addrinfo = socket.getaddrinfo(host, None)
                local_ip = None
                error = None
                for addr in addrinfo:
                    try:
                        local_ip = socket.gethostbyaddr(addr[4][0])
                        break
                    except (socket.gaierror, socket.herror,
                            socket.error) as err:
                        error = err

                if local_ip:
                    host_ip = ([local_ip[0]],
                               [fiveple[4][0] for fiveple in addrinfo],
                               [addrinfo[0][4][0]])
                else:
                    host_ip = ([], [], [])
                    if self.verbose:
                        print("WARNING: IP lookup by name failed for {0},"
                              "reason: {1}".format(host, error.strerror))
            extend_aliases(aliases, host_ip[1])
            extend_aliases(aliases, host_ip[2])

            return aliases

        def extend_aliases(aliases, als_new):
            """Does intersection of the two given list.
            Returns List of no repeated elements from the two list
            """
            if als_new:
                als_new = set(als_new)  # Create a set to remove duplicates.
                als_new = [alias for alias in als_new if (alias and
                                                          alias not in aliases
                                                          )]
                aliases.extend(als_new)

        host_or_ip = clean_IPv6(host_or_ip.lower())

        # for quickness, verify in the existing  aliases, if they are.
        if self.aliases:
            if host_or_ip.lower() in self.aliases:
                return True
            else:
                # get the alias for the given host_or_ip
                host_or_ip_aliases = get_aliases(host_or_ip)
                host_or_ip_aliases.append(host_or_ip)
                for alias in host_or_ip_aliases:
                    if alias in self.aliases:
                        #save the aliases for future
                        extend_aliases(self.aliases, host_or_ip_aliases)
                        return True
                return False

        # If not previews aliases save, look for them.
        # First, get the local information
        try:
            local_info = socket.gethostbyname_ex(socket.gethostname())
            local_aliases = [local_info[0].lower()]
            # if dotted host name, take first part and use as an alias
            try:
                local_aliases.append(local_info[0].split('.')[0])
            except:
                pass
            extend_aliases(local_aliases, ['127.0.0.1', 'localhost',
                                           '::1', '[::1]'])
            extend_aliases(local_aliases, local_info[1])
            extend_aliases(local_aliases, local_info[2])
            extend_aliases(local_aliases, get_aliases(socket.gethostname()))
        except (socket.herror, socket.gaierror, socket.error):
            # Try with the basic local aliases.
            local_aliases = ['127.0.0.1', 'localhost', '::1', '[::1]']

        # Get the aliases for this server host
        self.aliases = get_aliases(self.host)

        # List of possible suffixes
        suffixes = ('.local', '.lan',)

        # Check if this server is local
        for host in self.aliases:
            if host in local_aliases:
                # Is local then save the local aliases for future.
                extend_aliases(self.aliases, local_aliases)
                break
            # Handle special suffixes in hostnames.
            for suffix in suffixes:
                if host.endswith(suffix):
                    # Remove special suffix and attempt to match with local
                    # aliases.
                    host, _ = host.rsplit('.', 1)
                    if host in local_aliases:
                        # Is local then save the local aliases for future.
                        extend_aliases(self.aliases, local_aliases)
                        break

        # Check if the given host_or_ip is alias of the server host.
        if host_or_ip in self.aliases:
            return True

        # lastly get the alias for the given host_or_ip
        host_or_ip_aliases = get_aliases(host_or_ip)
        extend_aliases(host_or_ip_aliases, [host_or_ip])
        for alias in host_or_ip_aliases:
            if alias in self.aliases:
                #save the aliases for future
                extend_aliases(self.aliases, host_or_ip_aliases)
                return True
        return False

    def user_host_exists(self, user, host_or_ip):
        """Check to see if a user, host exists

        This method attempts to see if a user name matches the users on the
        server and that any user, host pair can match the host or IP address
        specified. This attempts to resolve wildcard matches.

        user[in]       user name
        host_or_ip[in] host or IP address

        Returns string - host from server that matches the host_or_ip or
                         None if no match.
        """
        res = self.exec_query("SELECT host FROM mysql.user WHERE user = '%s' "
                              "AND '%s' LIKE host " % (user, host_or_ip))
        if res:
            return res[0][0]
        return None

    def get_connection_values(self):
        """Return a dictionary of connection values for the server.

        Returns dictionary
        """
        conn_vals = {
            "user": self.user,
            "host": self.host
        }
        if self.passwd:
            conn_vals["passwd"] = self.passwd
        if self.socket:
            conn_vals["socket"] = self.socket
        if self.port:
            conn_vals["port"] = self.port

        return conn_vals

    def connect(self):
        """Connect to server

        Attempts to connect to the server as specified by the connection
        parameters.

        Note: This method must be called before executing queries.

        Raises UtilError if error during connect
        """
        try:
            self.db_conn = self.get_connection()
            # If no charset provided, get it from the "character_set_client"
            # server variable.
            if not self.charset:
                res = self.show_server_variable('character_set_client')
                self.db_conn.set_charset_collation(charset=res[0][1])
                self.charset = res[0][1]
        except UtilError:
            # Reset any previous value if the connection cannot be established,
            # before raising an exception. This prevents the use of a broken
            # database connection.
            self.db_conn = None
            raise
        self.connect_error = None
        self.read_only = self.show_server_variable("READ_ONLY")[0][1]

    def get_connection(self):
        """Return a new connection to the server.

        Attempts to connect to the server as specified by the connection
        parameters and returns a connection object.

        Return the resulting MySQL connection object or raises an UtilError if
        an error occurred during the server connection process.
        """
        try:
            parameters = {
                'user': self.user,
                'host': self.host,
                'port': self.port,
            }
            if self.socket and os.name == "posix":
                parameters['unix_socket'] = self.socket
            if self.passwd and self.passwd != "":
                parameters['passwd'] = self.passwd
            if self.charset:
                parameters['charset'] = self.charset
            parameters['host'] = parameters['host'].replace("[", "")
            parameters['host'] = parameters['host'].replace("]", "")
            db_conn = mysql.connector.connect(**parameters)
            # Return MySQL connection object.
            return db_conn
        except mysql.connector.Error as err:
            raise UtilError("Cannot connect to the {0} server.\n"
                            "Error {1}".format(self.role, err.msg), err.errno)

    def disconnect(self):
        """Disconnect from the server.
        """
        try:
            self.db_conn.disconnect()
        except:
            pass

    def get_version(self):
        """Return version number of the server.

        Get the server version. The respective instance variable is set with
        the result after querying the server the first time. The version is
        immediately returned when already known, avoiding querying the server
        at each time.

        Returns string - version string or None if error
        """
        # Return the local version value if already known.
        if self._version:
            return self._version

        # Query the server for its version.
        try:
            res = self.show_server_variable("VERSION")
            if res:
                self._version = res[0][1]
        except UtilError:
            # Ignore errors and return _version, initialized with None.
            pass

        return self._version

    def check_version_compat(self, t_major, t_minor, t_rel):
        """Checks version of the server against requested version.

        This method can be used to check for version compatibility.

        t_major[in]        target server version (major)
        t_minor[in]        target server version (minor)
        t_rel[in]          target server version (release)

        Returns bool True if server version is GE (>=) version specified,
                     False if server version is LT (<) version specified
        """
        version_str = self.get_version()
        if version_str is not None:
            match = re.match(r'^(\d+\.\d+(\.\d+)*).*$', version_str.strip())
            if match:
                version = [int(x) for x in match.group(1).split('.')]
                version = (version + [0])[:3]  # Ensure a 3 elements list
                return version >= [int(t_major), int(t_minor), int(t_rel)]
            else:
                return False
        return True

    def exec_query(self, query_str, options=None, exec_timeout=0):
        """Execute a query and return result set

        This is the singular method to execute queries. It should be the only
        method used as it contains critical error code to catch the issue
        with mysql.connector throwing an error on an empty result set.

        Note: will handle exception and print error if query fails

        Note: if fetchall is False, the method returns the cursor instance

        query_str[in]      The query to execute
        options[in]        Options to control behavior:
            params         Parameters for query
            columns        Add column headings as first row
                           (default is False)
            fetch          Execute the fetch as part of the operation and
                           use a buffered cursor
                           (default is True)
            raw            If True, use a buffered raw cursor
                           (default is True)
            commit         Perform a commit (if needed) automatically at the
                           end (default: True).
        exec_timeout[in]   Timeout value in seconds to kill the query execution
                           if exceeded. Value must be greater than zero for
                           this feature to be enabled. By default 0, meaning
                           that the query will not be killed.

        Returns result set or cursor
        """
        if options is None:
            options = {}
        params = options.get('params', ())
        columns = options.get('columns', False)
        fetch = options.get('fetch', True)
        raw = options.get('raw', True)
        do_commit = options.get('commit', True)

        # Guard for connect() prerequisite
        assert self.db_conn, "You must call connect before executing a query."

        # If we are fetching all, we need to use a buffered
        if fetch:
            if raw:
                cur = self.db_conn.cursor(
                    cursor_class=mysql.connector.cursor.MySQLCursorBufferedRaw)
            else:
                cur = self.db_conn.cursor(buffered=True)
        else:
            cur = self.db_conn.cursor(raw=True)

        # Execute query, handling parameters.
        q_killer = None
        try:
            if exec_timeout > 0:
                # Spawn thread to kill query if timeout is reached.
                # Note: set it as daemon to avoid waiting for it on exit.
                q_killer = QueryKillerThread(self, query_str, exec_timeout)
                q_killer.daemon = True
                q_killer.start()
            # Execute query.
            if params == ():
                cur.execute(query_str)
            else:
                cur.execute(query_str, params)
        except mysql.connector.Error as err:
            cur.close()
            if err.errno == CR_SERVER_LOST and exec_timeout > 0:
                # If the connection is killed (because the execution timeout is
                # reached), then it attempts to re-establish it (to execute
                # further queries) and raise a specific exception to track this
                # event.
                # CR_SERVER_LOST = Errno 2013 Lost connection to MySQL server
                # during query.
                self.db_conn.reconnect()
                raise UtilError("Timeout executing query", err.errno)
            else:
                raise UtilDBError("Query failed. {0}".format(err))
        except Exception:
            cur.close()
            raise UtilError("Unknown error. Command: {0}".format(query_str))
        finally:
            # Stop query killer thread if alive.
            if q_killer and q_killer.is_alive():
                q_killer.stop()

        # Fetch rows (only if available or fetch = True).
        if cur.with_rows:
            if fetch or columns:
                try:
                    results = cur.fetchall()
                    if columns:
                        col_headings = cur.column_names
                        col_names = []
                        for col in col_headings:
                            col_names.append(col)
                        results = col_names, results
                except mysql.connector.Error as err:
                    raise UtilDBError("Error fetching all query data: "
                                      "{0}".format(err))
                finally:
                    cur.close()
                return results
            else:
                # Return cursor to fetch rows elsewhere (fetch = false).
                return cur
        else:
            # No results (not a SELECT)
            try:
                if do_commit:
                    self.db_conn.commit()
            except mysql.connector.Error as err:
                raise UtilDBError("Error performing commit: {0}".format(err))
            finally:
                cur.close()
            return cur

    def commit(self):
        """Perform a COMMIT.
        """
        # Guard for connect() prerequisite
        assert self.db_conn, "You must call connect before executing a query."

        self.db_conn.commit()

    def rollback(self):
        """Perform a ROLLBACK.
        """
        # Guard for connect() prerequisite
        assert self.db_conn, "You must call connect before executing a query."

        self.db_conn.rollback()

    def show_server_variable(self, variable):
        """Returns one or more rows from the SHOW VARIABLES command.

        variable[in]       The variable or wildcard string

        Returns result set
        """

        return self.exec_query("SHOW VARIABLES LIKE '%s'" % variable)

    def get_uuid(self):
        """Return the uuid for this server if it is GTID aware.

        Returns uuid or None if server is not GTID aware.
        """
        if self.supports_gtid() != "NO":
            res = self.show_server_variable("server_uuid")
            return res[0][1]
        return None

    def supports_gtid(self):
        """Determine if server supports GTIDs

        Returns string - 'ON' = gtid supported and turned on,
                         'OFF' = supported but not enabled,
                         'NO' = not supported
        """
        # Check servers for GTID support
        version_ok = self.check_version_compat(5, 6, 5)
        if not version_ok:
            return "NO"
        try:
            res = self.exec_query("SELECT @@GLOBAL.GTID_MODE")
        except:
            return "NO"

        return res[0][0]

    def check_gtid_version(self):
        """Determine if server supports latest GTID changes

        This method checks the server to ensure it contains the latest
        changes to the GTID variables (from version 5.6.9).

        Raises UtilRplError when errors occur.
        """
        errors = []
        if not self.supports_gtid() == "ON":
            errors.append("    GTID is not enabled.")
        if not self.check_version_compat(5, 6, 9):
            errors.append("    Server version must be 5.6.9 or greater.")
        res = self.exec_query("SHOW VARIABLES LIKE 'gtid_executed'")
        if res == [] or not res[0][0] == "gtid_executed":
            errors.append("    Missing gtid_executed system variable.")
        if errors:
            errors = "\n".join(errors)
            errors = "\n".join([_GTID_ERROR % (self.host, self.port), errors])
            raise UtilRplError(errors)

    def check_gtid_executed(self, operation="copy"):
        """Check to see if the gtid_executed variable is clear

        If the value is not clear, raise an error with appropriate instructions
        for the user to correct the issue.

        operation[in]  Name of the operation (copy, import, etc.)
                       default = copy
        """
        res = self.exec_query("SHOW GLOBAL VARIABLES LIKE 'gtid_executed'")[0]
        if res[1].strip() == '':
            return
        err = ("The {0} operation contains GTID statements "
               "that require the global gtid_executed system variable on the "
               "target to be empty (no value). The gtid_executed value must "
               "be reset by issuing a RESET MASTER command on the target "
               "prior to attempting the {0} operation. "
               "Once the global gtid_executed value is cleared, you may "
               "retry the {0}.").format(operation)
        raise UtilRplError(err)

    def get_gtid_executed(self, skip_gtid_check=True):
        """Get the executed GTID set of the server.

        This function retrieves the (current) GTID_EXECUTED set of the server.

        skip_gtid_check[in]     Flag indicating if the check for GTID support
                                will be skipped or not. By default 'True'
                                (check is skipped).

        Returns a string with the GTID_EXECUTED set for this server.
        """
        if not skip_gtid_check:
            # Check server for GTID support.
            gtid_support = self.supports_gtid() == "NO"
            if gtid_support == 'NO':
                raise UtilRplError("Global Transaction IDs are not supported.")
            elif gtid_support == 'OFF':
                raise UtilError("Global Transaction IDs are not enabled.")
        # Get GTID_EXECUTED.
        try:
            return self.exec_query("SELECT @@GLOBAL.GTID_EXECUTED")[0][0]
        except UtilError:
            if skip_gtid_check:
                # Query likely failed because GTIDs are not supported,
                # therefore skip error in this case.
                return ""
            else:
                # If GTID check is not skipped re-raise exception.
                raise
        except IndexError:
            # If no rows are returned by query then return an empty string.
            return ''

    def gtid_subtract(self, gtid_set, gtid_subset):
        """Subtract given GTID sets.

        This function invokes GTID_SUBTRACT function on the server to retrieve
        the GTIDs from the given gtid_set that are not in the specified
        gtid_subset.

        gtid_set[in]        Base GTID set to subtract the subset from.
        gtid_subset[in]     GTID subset to be subtracted from the base set.

        Return a string with the GTID set resulting from the subtraction of the
        specified gtid_subset from the gtid_set.
        """
        try:
            return self.exec_query(
                "SELECT GTID_SUBTRACT('{0}', '{1}')".format(gtid_set,
                                                            gtid_subset)
            )[0][0]
        except IndexError:
            # If no rows are returned by query then return an empty string.
            return ''

    def gtid_subtract_executed(self, gtid_set):
        """Subtract GTID_EXECUTED to the given GTID set.

        This function invokes GTID_SUBTRACT function on the server to retrieve
        the GTIDs from the given gtid_set that are not in the GTID_EXECUTED
        set.

        gtid_set[in]        Base GTID set to subtract the GTID_EXECUTED.

        Return a string with the GTID set resulting from the subtraction of the
        GTID_EXECUTED set from the specified gtid_set.
        """
        from mysql.utilities.common.topology import _GTID_SUBTRACT_TO_EXECUTED
        try:
            return self.exec_query(
                _GTID_SUBTRACT_TO_EXECUTED.format(gtid_set)
            )[0][0]
        except IndexError:
            # If no rows are returned by query then return an empty string.
            return ''

    def checksum_table(self, tbl_name, exec_timeout=0):
        """Compute checksum of specified table (CHECKSUM TABLE tbl_name).

        This function executes the CHECKSUM TABLE statement for the specified
        table and returns the result. The CHECKSUM is aborted (query killed)
        if a timeout value (greater than zero) is specified and the execution
        takes longer than the specified time.

        tbl_name[in]        Name of the table to perform the checksum.
        exec_timeout[in]    Maximum execution time (in seconds) of the query
                            after which it will be killed. By default 0, no
                            timeout.

        Returns a tuple with the checksum result for the target table. The
        first tuple element contains the result from the CHECKSUM TABLE query
        or None if an error occurred (e.g. execution timeout reached). The
        second element holds any error message or None if the operation was
        successful.
        """
        try:
            return self.exec_query(
                "CHECKSUM TABLE {0}".format(tbl_name),
                exec_timeout=exec_timeout
            )[0], None
        except IndexError:
            # If no rows are returned by query then return None.
            return None, "No data returned by CHECKSUM TABLE"
        except UtilError as err:
            # Return None if the query is killed (exec_timeout reached).
            return None, err.errmsg

    def get_gtid_status(self):
        """Get the GTID information for the server.

        This method attempts to retrieve the GTID lists. If the server
        does not have GTID turned on or does not support GTID, the method
        will throw and exception.

        Returns [list, list, list]
        """
        # Check servers for GTID support
        if self.supports_gtid() == "NO":
            raise UtilError("Global Transaction IDs are not supported.")

        res = self.exec_query("SELECT @@GLOBAL.GTID_MODE")
        if res[0][0].upper() == 'OFF':
            raise UtilError("Global Transaction IDs are not enabled.")

        gtid_data = [self.exec_query("SELECT @@GLOBAL.GTID_EXECUTED")[0],
                     self.exec_query("SELECT @@GLOBAL.GTID_PURGED")[0],
                     self.exec_query("SELECT @@GLOBAL.GTID_OWNED")[0]]

        return gtid_data

    def check_rpl_user(self, user, host):
        """Check replication user exists and has the correct privileges.

        user[in]      user name of rpl_user
        host[in]      host name of rpl_user

        Returns [] - no exceptions, list if exceptions found
        """
        errors = []
        ipv6 = False
        if "]" in host:
            ipv6 = True
            host = clean_IPv6(host)
        result = self.user_host_exists(user, host)
        if ipv6:
            result = format_IPv6(result)
        if result is None or result == []:
            errors.append("The replication user %s@%s was not found "
                          "on %s:%s." % (user, host, self.host, self.port))
        else:
            rpl_user = User(self, "%s@" % user + result)
            if not rpl_user.has_privilege('*', '*',
                                          'REPLICATION SLAVE'):
                errors.append("Replication user does not have the "
                              "correct privilege. She needs "
                              "'REPLICATION SLAVE' on all replicated "
                              "databases.")

        return errors

    def supports_plugin(self, plugin):
        """Check if the given plugin is supported.

        Check to see if the server supports a plugin. Return True if
        plugin installed and active.

        plugin[in]     Name of plugin to check

        Returns True if plugin is supported, and False otherwise.
        """
        _PLUGIN_QUERY = ("SELECT * FROM INFORMATION_SCHEMA.PLUGINS "
                         "WHERE PLUGIN_NAME ")
        res = self.exec_query("".join([_PLUGIN_QUERY, "LIKE ",
                                       "'%s" % plugin, "%'"]))
        if not res:
            return False
        # Now see if it is active.
        elif res[0][2] != 'ACTIVE':
            return False
        return True

    def get_all_databases(self):
        """Return a result set containing all databases on the server
        except for internal databases (mysql, INFORMATION_SCHEMA,
        PERFORMANCE_SCHEMA)

        Returns result set
        """

        _GET_DATABASES = """
        SELECT SCHEMA_NAME
        FROM INFORMATION_SCHEMA.SCHEMATA
        WHERE SCHEMA_NAME != 'INFORMATION_SCHEMA'
        AND SCHEMA_NAME != 'PERFORMANCE_SCHEMA'
        AND SCHEMA_NAME != 'mysql'
        """
        return self.exec_query(_GET_DATABASES)

    def get_storage_engines(self):
        """Return list of storage engines on this server.

        Returns (list) (engine, support, comment)
        """

        _QUERY = """
            SELECT UPPER(engine), UPPER(support)
            FROM INFORMATION_SCHEMA.ENGINES
            ORDER BY engine
        """
        return self.exec_query(_QUERY)

    def check_storage_engines(self, other_list):
        """Compare storage engines from another server.

        This method compares the list of storage engines for the current
        server against a list supplied as **other_list**. It returns two
        lists - one for the storage engines on this server not on the other
        list, and another for the storage engines on the other list not on this
        server.

        Note: type case sensitive - make sure list is in uppercase

        other_list[in]     A list from another server in the form
                           (engine, support) - same output as
                           get_storage_engines()

        Returns (list, list)
        """
        # Guard for connect() prerequisite
        assert self.db_conn, "You must call connect before check engine lists."

        def _convert_set_to_list(set_items):
            """Convert a set to list
            """
            if len(set_items) > 0:
                item_list = []
                for item in set_items:
                    item_list.append(item)
            else:
                item_list = None
            return item_list

        # trivial, but guard against misuse
        this_list = self.get_storage_engines()
        if other_list is None:
            return (this_list, None)

        same = set(this_list) & set(other_list)
        master_extra = _convert_set_to_list(set(this_list) - same)
        slave_extra = _convert_set_to_list(set(other_list) - same)

        return (master_extra, slave_extra)

    def has_storage_engine(self, target):
        """Check to see if an engine exists and is supported.

        target[in]     name of engine to find

        Returns bool True - engine exists and is active, false = does not
                     exist or is not supported/not active/disabled
        """
        if len(target) == 0:
            return True  # This says we will use default engine on the server.
        if target is not None:
            engines = self.get_storage_engines()
            for engine in engines:
                if engine[0].upper() == target.upper() and \
                   engine[1].upper() in ['YES', 'DEFAULT']:
                    return True
        return False

    def substitute_engine(self, tbl_name, create_str,
                          new_engine, def_engine, quiet=False):
        """Replace storage engine in CREATE TABLE

        This method will replace the storage engine in the CREATE statement
        under the following conditions:
            - If new_engine is specified and it exists on destination, use it.
            - Else if existing engine does not exist and def_engine is specfied
              and it exists on destination, use it. Also, don't substitute if
              the existing engine will not be changed.

        tbl_name[in]       table name
        create_str[in]     CREATE statement
        new_engine[in]     name of storage engine to substitute (convert to)
        def_engine[in]     name of storage engine to use if existing engines
                           does not exist

        Returns string CREATE string with replacements if found, else return
                       original string
        """
        res = [create_str]
        exist_engine = ''
        is_create_like = False
        replace_msg = "# Replacing ENGINE=%s with ENGINE=%s for table %s."
        add_msg = "# Adding missing ENGINE=%s clause for table %s."
        if new_engine is not None or def_engine is not None:
            i = create_str.find("ENGINE=")
            if i > 0:
                j = create_str.find(" ", i)
                exist_engine = create_str[i + 7:j]
            else:
                ## Check if it is a CREATE TABLE LIKE statement
                is_create_like = (create_str.find("CREATE TABLE {0} LIKE"
                                                  "".format(tbl_name)) == 0)

        # Set default engine
        #
        # If a default engine is specified and is not the same as the
        # engine specified in the table CREATE statement (existing engine) if
        # specified, and both engines exist on the server, replace the existing
        # engine with the default engine.
        #
        if def_engine is not None and \
                exist_engine.upper() != def_engine.upper() and \
                self.has_storage_engine(def_engine) and \
                self.has_storage_engine(exist_engine):

            # If no ENGINE= clause present, add it
            if len(exist_engine) == 0:
                if is_create_like:
                    alter_str = "ALTER TABLE {0} ENGINE={1}".format(tbl_name,
                                                                    def_engine)
                    res = [create_str, alter_str]
                else:
                    i = create_str.find(";")
                    i = len(create_str) if i == -1 else i
                    create_str = "{0} ENGINE={1};".format(create_str[0:i],
                                                          def_engine)
                    res = [create_str]
            # replace the existing storage engine
            else:
                create_str.replace("ENGINE=%s" % exist_engine,
                                   "ENGINE=%s" % def_engine)
            if not quiet:
                if len(exist_engine) > 0:
                    print replace_msg % (exist_engine, def_engine, tbl_name)
                else:
                    print add_msg % (def_engine, tbl_name)
            exist_engine = def_engine

        # Use new engine
        if (new_engine is not None and
                exist_engine.upper() != new_engine.upper() and
                self.has_storage_engine(new_engine)):
            if len(exist_engine) == 0:
                if is_create_like:
                    alter_str = "ALTER TABLE {0} ENGINE={1}".format(tbl_name,
                                                                    new_engine)
                    res = [create_str, alter_str]
                else:
                    i = create_str.find(";")
                    i = len(create_str) if i == -1 else i
                    create_str = "{0} ENGINE={1};".format(create_str[0:i],
                                                          new_engine)
                    res = [create_str]
            else:
                create_str = create_str.replace("ENGINE=%s" % exist_engine,
                                                "ENGINE=%s" % new_engine)
                res = [create_str]
            if not quiet:
                if len(exist_engine) > 0:
                    print replace_msg % (exist_engine, new_engine, tbl_name)
                else:
                    print add_msg % (new_engine, tbl_name)
        return res

    def get_innodb_stats(self):
        """Return type of InnoDB engine and its version information.

        This method returns a tuple containing the type of InnoDB storage
        engine (builtin or plugin) and the version number reported.

        Returns (tuple) (type = 'builtin' or 'plugin', version_number,
                         have_innodb = True or False)
        """
        # Guard for connect() prerequisite
        assert self.db_conn, "You must call connect before get innodb stats."

        _BUILTIN = """
            SELECT (support='YES' OR support='DEFAULT' OR support='ENABLED')
            AS `exists` FROM INFORMATION_SCHEMA.ENGINES
            WHERE engine = 'innodb';
        """
        _PLUGIN = """
            SELECT (plugin_library LIKE 'ha_innodb_plugin%') AS `exists`
            FROM INFORMATION_SCHEMA.PLUGINS
            WHERE LOWER(plugin_name) = 'innodb' AND
                  LOWER(plugin_status) = 'active';
        """
        _VERSION = """
            SELECT plugin_version, plugin_type_version
            FROM INFORMATION_SCHEMA.PLUGINS
            WHERE LOWER(plugin_name) = 'innodb';
        """

        inno_type = None
        results = self.exec_query(_BUILTIN)
        if results is not None and results != () and results[0][0] is not None:
            inno_type = "builtin"

        results = self.exec_query(_PLUGIN)
        if results is not None and results != () and \
           results != [] and results[0][0] is not None:
            inno_type = "plugin "

        results = self.exec_query(_VERSION)
        version = []
        if results is not None:
            version.append(results[0][0])
            version.append(results[0][1])
        else:
            version.append(None)
            version.append(None)

        results = self.show_server_variable("have_innodb")
        if results is not None and results != [] and \
           results[0][1].lower() == "yes":
            have_innodb = True
        else:
            have_innodb = False

        return (inno_type, version[0], version[1], have_innodb)

    def read_and_exec_SQL(self, input_file, verbose=False):
        """Read an input file containing SQL statements and execute them.

        input_file[in]     The full path to the file
        verbose[in]        Print the command read
                           Default = False

        Returns True = success, False = error

        TODO : Make method read multi-line queries.
        """
        f_input = open(input_file)
        res = True
        while True:
            cmd = f_input.readline()
            if not cmd:
                break
            res = None
            if len(cmd) > 1:
                if cmd[0] != '#':
                    if verbose:
                        print cmd
                    query_options = {
                        'fetch': False
                    }
                    res = self.exec_query(cmd, query_options)
        f_input.close()
        return res

    def binlog_enabled(self):
        """Check binary logging status for the client.

        Returns bool - True - binary logging is ON, False = OFF
        """
        res = self.show_server_variable("log_bin")
        if not res:
            raise UtilRplError("Cannot retrieve status of log_bin variable.")
        if res[0][1] in ("OFF", "0"):
            return False
        return True

    def toggle_binlog(self, action="disable"):
        """Enable or disable binary logging for the client.

        Note: user must have SUPER privilege

        action[in]         if 'disable', turn off the binary log
                           elif 'enable' turn binary log on
                           do nothing if action != 'enable' or 'disable'
        """

        if action.lower() == 'disable':
            self.exec_query("SET SQL_LOG_BIN=0")
        elif action.lower() == 'enable':
            self.exec_query("SET SQL_LOG_BIN=1")

    def foreign_key_checks_enabled(self, force=False):
        """Check foreign key status for the connection.
        force[in]       if True, returns the value directly from the server
                        instead of returning the cached fkey value

        Returns bool - True - foreign keys are enabled
        """
        if self.fkeys is None or force:
            res = self.show_server_variable("foreign_key_checks")
            self.fkeys = (res is not None) and (res[0][1] == "ON")
        return self.fkeys

    def disable_foreign_key_checks(self, disable=True):
        """Enable or disable foreign key checks for the connection.

        disable[in]     if True, turn off foreign key checks
                        elif False turn foreign key checks on.
        """
        if self.fkeys is None:
            self.foreign_key_checks_enabled()

        # Only do something if foreign keys are OFF and shouldn't be disabled
        # or if they are ON and should be disabled
        if self.fkeys == disable:
            val = "OFF" if disable else "ON"
            self.exec_query(_FOREIGN_KEY_SET.format(val),
                            {'fetch': False, 'commit': False})
            self.fkeys = not self.fkeys

    def autocommit_set(self):
        """Check autocommit status for the connection.

        Returns bool - True if autocommit is enabled and False otherwise.
        """
        if self.autocommit is None:
            res = self.show_server_variable('autocommit')
            self.autocommit = (res and res[0][1] == '1')
        return self.autocommit

    def toggle_autocommit(self, enable=None):
        """Enable or disable autocommit for the connection.

        This method switch the autocommit value or enable/disable it according
        to the given parameter.

        enable[in]         if True, turn on autocommit (set to 1)
                           else if False turn autocommit off (set to 0).
        """
        if enable is None:
             # Switch autocommit value.
            if self.autocommit is None:
                # Get autocommit value if unknown
                self.autocommit_set()
            if self.autocommit:
                value = '0'
                self.autocommit = False
            else:
                value = '1'
                self.autocommit = True
        else:
            # Set AUTOCOMMIT according to provided value.
            if enable:
                value = '1'
                self.autocommit = True
            else:
                value = '0'
                self.autocommit = False
        # Change autocommit value.
        self.exec_query(_AUTOCOMMIT_SET.format(value), {'fetch': 'false'})

    def get_server_id(self):
        """Retrieve the server id.

        Returns int - server id.
        """
        try:
            res = self.show_server_variable("server_id")
        except:
            raise UtilRplError("Cannot retrieve server id from "
                               "%s." % self.role)

        return int(res[0][1])

    def get_server_uuid(self):
        """Retrieve the server uuid.

        Returns string - server uuid.
        """
        try:
            res = self.show_server_variable("server_uuid")
            if res is None or res == []:
                return None
        except:
            raise UtilRplError("Cannot retrieve server_uuid from "
                               "%s." % self.role)

        return res[0][1]

    def get_lctn(self):
        """Get lower_case_table_name setting.

        Returns lctn value or None if cannot get value
        """
        res = self.show_server_variable("lower_case_table_names")
        if res != []:
            return res[0][1]
        return None

    def get_binary_logs(self, options=None):
        """Return a list of the binary logs.

        options[in]        query options

        Returns list - binlogs or None if binary logging turned off
        """
        if options is None:
            options = {}
        if self.binlog_enabled():
            return self.exec_query("SHOW BINARY LOGS", options)

        return None

    def set_read_only(self, on=False):
        """Turn read only mode on/off

        on[in]         if True, turn read_only ON
                       Default is False
        """
        # Only turn on|off read only if it were off at connect()
        if not self.read_only:
            return self.exec_query("SET @@GLOBAL.READ_ONLY = %s" %
                                   "ON" if on else "OFF")
        return None

    def grant_tables_enabled(self):
        """Check to see if grant tables are enabled

        Returns bool - True = grant tables are enabled, False = disabled
        """
        if self.grants_enabled is None:
            try:
                self.exec_query("SHOW GRANTS FOR 'snuffles'@'host'")
                self.grants_enabled = True
            except UtilError as error:
                if "--skip-grant-tables" in error.errmsg:
                    self.grants_enabled = False
                # Ignore other errors as they are not pertinent to the check
                else:
                    self.grants_enabled = True
        return self.grants_enabled


class QueryKillerThread(threading.Thread):
    """Class to run a thread to kill an executing query.

    This class is used to spawn a thread than will kill the execution
    (connection) of a query upon reaching a given timeout.
    """

    def __init__(self, server, query, timeout):
        """Constructor.

        server[in]      Server instance where the target query is executed.
        query[in]       Target query to kill.
        timeout[in]     Timeout value in seconds used to kill the query when
                        reached.
        """
        threading.Thread.__init__(self)
        self._stop_event = threading.Event()
        self._query = query
        self._timeout = timeout
        self._server = server
        self._connection = server.get_connection()
        server.get_version()

    def run(self):
        """Main execution of the query killer thread.
        Stop the thread if instructed as such
        """
        connector_error = None
        # Kill the query connection upon reaching the given execution timeout.
        while not self._stop_event.is_set():
            # Wait during the defined time.
            self._stop_event.wait(self._timeout)
            # If the thread was asked to stop during wait, it does not try to
            # kill the query.
            if not self._stop_event.is_set():
                try:
                    cur = self._connection.cursor(raw=True)

                    # Get process information from threads table when available
                    # (for versions > 5.6.1), since it does not require a mutex
                    # and has minimal impact on server performance.
                    if self._server.check_version_compat(5, 6, 1):
                        cur.execute(
                            "SELECT processlist_id "
                            "FROM performance_schema.threads"
                            " WHERE processlist_command='Query'"
                            " AND processlist_info='{0}'".format(self._query))
                    else:
                        cur.execute(
                            "SELECT id FROM information_schema.processlist"
                            " WHERE command='Query'"
                            " AND info='{0}'".format(self._query))
                    result = cur.fetchall()

                    try:
                        process_id = result[0][0]
                    except IndexError:
                        # No rows are returned if the query ended in the
                        # meantime.
                        process_id = None

                    # Kill the connection associated to que process id.
                    # Note: killing the query will not work with
                    # connector-python,since it will hang waiting for the
                    #  query to return.
                    if process_id:
                        cur.execute("KILL {0}".format(process_id))
                except mysql.connector.Error as err:
                    # Hold error to raise at the end.
                    connector_error = err
                finally:
                    # Close cursor if available.
                    if cur:
                        cur.close()
                # Stop this thread.
                self.stop()

        # Close connection.
        try:
            self._connection.disconnect()
        except mysql.connector.Error:
            # Only raise error if no previous error has occurred.
            if not connector_error:
                raise
        finally:
            # Raise any previous error that already occurred.
            if connector_error is not None:
                # pylint: disable=E0702
                raise connector_error

    def stop(self):
        """Stop the thread.

        Set the event flag for the thread to stop as soon as possible.
        """
        self._stop_event.set()



#
# Copyright (c) 2010, 2013, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This module contains an abstraction of a topolgy map object used to discover
slaves and down-stream replicants for mapping topologies.
"""

import getpass

from mysql.utilities.common.options import parse_user_password
from mysql.utilities.common.replication import Slave
from mysql.utilities.common.server import connect_servers
from mysql.utilities.common.user import User
from mysql.utilities.exception import UtilError


_START_PORT = 3306


class TopologyMap(object):
    """The TopologyMap class can be used to connect to a running MySQL server
    and discover its slaves. Setting the option "recurse" permits the
    class to discover a replication topology by finding the slaves for each
    slave for the first master requested.

    To generate a topology map, the caller must call the
    generate_topology_map() method to build the topology. This is left as a
    separate state because it can be a lengthy process thereby too long for a
    constructor method.

    The class also includes methods for printing a graph of the topology
    as well as returning a list of master, slave tuples reporting the
    host name and port for each.
    """

    def __init__(self, seed_server, options=None):
        """Constructor

        seed_server[in]    Master (seed) server connection dictionary
        options[in]        options for controlling behavior:
          recurse          If True, check each slave found for add'l slaves
                           Default = False
          prompt_user      If True, prompt user if slave connection fails with
                           master connection parameters
                           Default = False
          quiet            if True, print only the data
                           Default = False
          width            width of report
                           Default = 75
          num_retries      Number of times to retry a failed connection attempt
                           Default = 0
        """
        if options is None:
            options = {}
        self.recurse = options.get("recurse", False)
        self.quiet = options.get("quiet", False)
        self.prompt_user = options.get("prompt", False)
        self.num_retries = options.get("num_retries", 0)
        self.socket_path = options.get("socket_path", None)
        self.verbose = options.get('verbosity', 0) > 0
        self.seed_server = seed_server
        self.topology = []
        self.options = options

    def _connect(self, conn):
        """Find the attached slaves for a list of server connections.

        This method connects to each server in the list and retrieves its
        slaves.
        It can be called recursively if the recurse parameter is True.

        conn[in]           Connection dictionary used to connect to server

        Returns tuple - master Server class instance, master:host string
        """
        conn_options = {
            'quiet': self.quiet,
            'src_name': "master",
            'dest_name': None,
            'version': "5.0.0",
            'unique': True,
        }

        master_info = "%s:%s" % (conn['host'],
                                 conn['port'])
        master = None

        # Clear socket if used with a local server
        if (conn['host'] == 'localhost' or conn['host'] == "127.0.0.1" or
           conn['host'] == "::1" or conn['host'] == "[::1]"):
            conn['unix_socket'] = None

        # Increment num_retries if not set when --prompt is used
        if self.prompt_user and self.num_retries == 0:
            self.num_retries += 1

        # Attempt to connect to the server given the retry limit
        for i in range(0, self.num_retries + 1):
            try:
                servers = connect_servers(conn, None, conn_options)
                master = servers[0]
                break
            except UtilError, e:
                print "FAILED.\n"
                if i < self.num_retries and self.prompt_user:
                    print "Connection to %s has failed.\n" % master_info + \
                        "Please enter the following information " + \
                        "to connect to this server."
                    conn['user'] = raw_input("User name: ")
                    conn['passwd'] = getpass.getpass("Password: ")
                else:
                    # retries expired - re-raise error if still failing
                    raise UtilError(e.errmsg)

        return (master, master_info)

    @staticmethod
    def _check_permissions(server, priv):
        """Check to see if user has permissions to execute.

        server[in]     Server class instance
        priv[in]       privilege to check

        Returns True if permissions available, raises exception if not
        """
        # Check user permissions
        user_pass_host = server.user
        if server.passwd is not None and len(server.passwd) > 0:
            user_pass_host += ":" + server.passwd
        user_pass_host += "@" + server.host
        user = User(server, user_pass_host, False)
        if not user.has_privilege("*", "*", priv):
            raise UtilError("Not enough permissions. The user must have the "
                            "%s privilege." % priv)

    def _get_slaves(self, max_depth, seed_conn=None, masters_found=None):
        """Find the attached slaves for a list of server connections.

        This method connects to each server in the list and retrieves its
        slaves. It can be called recursively if the recurse option is True.

        max_depth[in]       Maximum depth of recursive search
        seed_conn[in]       Current master connection dictionary. Initially,
                            this is the seed server (original master defined
                            in constructor)
        masters_found[in]   a list of all servers in master roles - used to
                            detect a circular replication topology. Initially,
                            this is an empty list as the master detection must
                            occur as the topology is traversed.

        Returns list - list of slaves connected to each server in list
        """
        if not masters_found:
            masters_found = []
        topology = []
        if seed_conn is None:
            seed_conn = self.seed_server

        master, master_info = self._connect(seed_conn)
        if master is None:
            return []

        # Check user permissions
        self._check_permissions(master, "REPLICATION SLAVE")

        # Save the master for circular replication identification
        masters_found.append(master_info)

        if not self.quiet:
            print "# Finding slaves for master: %s" % master_info

        # See if the user wants us to discover slaves.
        discover = self.options.get("discover", None)
        if discover is None:
            return

        # Get user and password (supports login-paths)
        user, password = parse_user_password(discover, options=self.options)

        # Get replication topology
        slaves = master.get_slaves(user, password)
        slave_list = []
        depth = 0
        if len(slaves) > 0:
            for slave in slaves:
                if slave.find(":") > 0:
                    host, port = slave.split(":", 1)
                else:
                    host = slave
                    port = _START_PORT  # Use the default
                slave_conn = self.seed_server.copy()
                slave_conn['host'] = host
                slave_conn['port'] = port

                io_sql_running = None
                # If verbose then get slave threads (IO and SQL) status
                if self.verbose:
                    # Create slave instance
                    conn_dict = {
                        'conn_info': {'user': user, 'passwd': password,
                                      'host': host, 'port': port,
                                      'socket': None},
                        'role': slave,
                        'verbose': self.verbose
                    }
                    slave_obj = Slave(conn_dict)
                    # Get IO and SQL status
                    try:
                        slave_obj.connect()
                        thread_status = slave_obj.get_thread_status()
                        if thread_status:
                            io_sql_running = (thread_status[1],
                                              thread_status[2])
                    except UtilError:
                        # Connection error
                        io_sql_running = ('ERROR', 'ERROR')

                # Now check for circular replication topology - do not recurse
                # if slave is also a master.
                if self.recurse and not slave in masters_found and \
                   ((max_depth is None) or (depth < max_depth)):
                    new_list = self._get_slaves(max_depth, slave_conn,
                                                masters_found)
                    if new_list == []:
                        slave_list.append((slave, [], io_sql_running))
                    else:
                        # Add IO and SQL state to slave from recursion
                        if io_sql_running:
                            new_list = [(new_list[0][0], new_list[0][1],
                                         io_sql_running)]
                        slave_list.append(new_list)
                    depth += 1
                else:
                    slave_list.append((slave, [], io_sql_running))
        topology.append((master_info, slave_list))

        return topology

    def generate_topology_map(self, max_depth):
        """Find the attached slaves for a list of server connections.

        This method generates the topology for the seed server specified at
        instantiation.

        max_depth[in]       Maximum depth of recursive search
        """
        self.topology = self._get_slaves(max_depth)

    def depth(self):
        """Return depth of the topology tree.

        Returns int - depth of topology tree.
        """
        return len(self.topology)

    def slaves_found(self):
        """Check to see if any slaves were found.

        Returns bool - True if slaves found, False if no slaves.
        """
        return not (len(self.topology) and self.topology[0][1] == [])

    def print_graph(self, topology_list=None, masters_found=None,
                    level=0, preamble=""):
        """Prints a graph of the topology map to standard output.

        This method traverses a list of the topology and prints a graph. The
        method is designed to be recursive traversing the list to print the
        slaves for each master in the topology. It will also detect a circular
        replication segment and indicate it on the graph.

        topology_list[in]   a list in the form (master, slave) of server
        masters_found[in]   a list of all servers in master roles - used to
                            detect a circular replication topology. Initially,
                            this is an empty list as the master detection must
                            occur as the topology is traversed.
        level[in]           the level of indentation - increases with each
                            set of slaves found in topology
        preamble[in]        prefix calculated during recursion to indent text
        """
        if not topology_list:
            topology_list = []
        if not masters_found:
            masters_found = []
        # if first iteration, use the topology list generated earlier
        if topology_list == []:
            if self.topology == []:
                # topology not generated yet
                raise UtilError("You must first generate the topology.")
            topology_list = self.topology

        # Detect if we are looking at a sublist or not. Get sublist.
        if len(topology_list) == 1:
            topology_list = topology_list[0]
        master = topology_list[0]

        # Save the master for circular replication identification
        masters_found.append(master)

        # For each slave, print the graph link
        slaves = topology_list[1]
        stop = len(slaves)
        if stop > 0:
            # Level 0 is always the first master in the topology.
            if level == 0:
                print("{0} (MASTER)".format(master))
            for i in range(0, stop):
                if len(slaves[i]) == 1:
                    slave = slaves[i][0]
                else:
                    slave = slaves[i]
                new_preamble = "{0}   ".format(preamble)
                print("{0}|".format(new_preamble))
                role = "(SLAVE"
                if not slave[1] == [] or slave[0] in masters_found:
                    role = "{0} + MASTER".format(role)
                role = "{0})".format(role)

                # Print threads (IO and SQL) status if verbose
                t_status = ''
                if self.verbose:
                    try:
                        t_status = " [IO: {0}, SQL: {1}]".format(slave[2][0],
                                                                 slave[2][1])
                    except IndexError:
                        # This should never happened... (done to avoid crash)
                        t_status = " [IO: ??, SQL: ??]"

                print "{0}+--- {1}{2}".format(new_preamble, slave[0],
                                              t_status),

                if (slave[0] in masters_found):
                    print "<-->",
                else:
                    print "-",
                print role

                if not slave[1] == []:
                    if i < stop - 1:
                        new_preamble = "{0}|".format(new_preamble)
                    else:
                        new_preamble = "{0} ".format(new_preamble)
                    self.print_graph(slave, masters_found,
                                     level + 1, new_preamble)

    def _get_row(self, topology_list):
        """Get a row (master, slave) for the topology map.

        topology_list[in]  The topology list

        Returns tuple - a row (master, slave)
        """
        new_row = []
        if len(topology_list) == 1:
            topology_list = topology_list[0]
        master = topology_list[0]
        slaves = topology_list[1]
        for slave in slaves:
            if len(slave) == 1:
                new_slave = slave[0]
            else:
                new_slave = slave
            new_row.append((master, new_slave[0]))
            new_row.extend(self._get_row(new_slave))
        return new_row

    def get_topology_map(self):
        """Get a list of the topology map suitable for export

        Returns list - a list of masters and their slaves in two columns
        """
        # Get a row for the list
        # make a list from the topology
        master_slaves = [self._get_row(row) for row in self.topology]
        return master_slaves[0]



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Errors raised within the MySQL Fabric library.
"""

class Error(Exception):
    """Base exception for all errors in the package.
    """
    pass

class NotCallableError(Error):
    """Exception raised when a callable was expected but not provided.
    """
    pass

class NotEventError(Error):
    """Exception raised when a non-event instance was passed where an
    event instance was expected.
    """
    pass

class UnknownCallableError(Error):
    """Exception raised when trying to use a callable that was not
    known when a known callable was expected.
    """
    pass

class ExecutorError(Error):
    """Exception raised when the one tries to access the executor that
    is not properly configured.
    """

class InvalidGtidError(Error):
    """Exception raised when the one tries to use and make operations with
    invalid GTID(s).
    """

class UuidError(Error):
    """Exception raised when there are problems with uuids. For example,
    if the expected uuid does not match the server's uuid.
    """
    pass

class TimeoutError(Error):
    """Exception raised when there is a timeout.
    """

class DatabaseError(Error):
    """Exception raised when something bad happens while accessing a
    database.
    """
    pass

class ProgrammingError(Error):
    """Exception raised when a developer tries to use the interfaces and
    executes an invalid operation.
    """
    pass

class ConfigurationError(ProgrammingError):
    """Exception raised when configuration options are not properly set.
    """
    pass

class LockManagerError(Error):
    """Exception raised when an invalid operation is attempted on the
    lock manager or locks are broken.
    """
    pass

class ServiceError(Error):
    """Exception raised when one tries to use the service interface and
    executes an invalid operation.
    """
    pass

class GroupError(ServiceError):
    """Exception raised when one tries to execute an invalid operation on a
    group. For example, it is not possible to create two groups with the
    same id or remove a group that has associated servers.
    """
    pass

class ServerError(ServiceError):
    """Exception raised when one tries to execute an invalid operation on a
    server. For example, it is not possible to create two servers with the
    same uuid.
    """
    pass

class ProcedureError(ServiceError):
    """Exception raised when a procedure is not found.
    """
    pass

class ShardingError(ServiceError):
    """Exception raised when an invalid operation is attempted on the
    sharding system.
    """
    pass

class BackupError(Error):
    """Exception raised when a error occurs in the backup restore framework
    of Fabric.
    """
    pass

class CredentialError(Error):
    """Exception raised when something is wrong with credentials"""
    pass



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Module containing classes and functions for working with the
persistent store.

The two main classes here are the :class:`MySQLPersister` and the
:class:`PersistentMeta`.

Initializing the persistence system
-----------------------------------

Before creating any persisters, it is necessary to set up the
persistance system using connection information where objects should
be persisted::

   import mysql.fabric.persistence as persistence

   persistence.init(host="localhost", port=3307,
                    user="fabric_user", password="xyzzy")


Thread Initialization
~~~~~~~~~~~~~~~~~~~~~

When starting a new thread, a persister have to be created.  Typical
code for initializing the persister system for a thread is::

   import mysql.fabric.persistence as persistence
   persistence.init_thread()

"""
import functools
import inspect
import logging
import threading
import time
import uuid as _uuid

import mysql.fabric.server_utils as _server_utils
import mysql.fabric.errors as _errors

DEFAULT_DATABASE = 'fabric'
DEFAULT_CONNECT_ATTEMPTS = 0
DEFAULT_CONNECT_DELAY = 0

class PersistentMeta(type):
    """Metaclass for persistent classes.

    This is a metaclass for persistent classes. The metaclass is
    responsible for:

    - Keep track of the persister for the thread. There is exactly one
      persister for each thread and a thread is expected call the
      :meth:`init_thread` method with the persister when it starts.

    - Add the 'persister' attribute to each method, giving it the
      persister class assigned to the thread by the object store.

    - Registering the class with the object store so that it will get
      an init() and deinit() call when the state store inits and
      deinits.
    """
    thread_local = threading.local() # Thread-local store
    classes = []                     # List of all persistent classes

    @classmethod
    def init_thread(mcs, persister):
        """Initialize thread-specific data.

        :param persister: Persister to use for the thread.
        """

        mcs.thread_local.persister = persister

    @classmethod
    def deinit_thread(mcs):
        """De-initialize thread-specific data.
        """

        mcs.thread_local.persister = None

    @classmethod
    def wrapfunc(mcs, func):
        """Wrap the function to pass the persister for the thread.

        The function is wrapped by adding the argument 'persister' if
        the function request it by having an explicit argument named
        'persister'.

        The wrapped function will set the 'persister' argument to the
        persister assigned to the thread, but if an explicit persister
        argument has been given in the call, it is passed.

        Note that the function is not wrapped if it has a keywords
        argument (** argument) even though it is possible to pass the
        persister to such a function.  The reason for this is that
        many function wrappers use keywords argument to pass the
        arguments to the underlying function.

        """

        # Check that the function can accept a 'persister' parameter
        # or accepts keywords.
        argspec = inspect.getargspec(func)
        if 'persister' not in argspec.args:
            return func         # No need to wrap it, so just return it

        original = func         # Create closure
        @functools.wraps(func)
        def _wrap(*args, **kwrds):
            """Inner wrapper function.
            """
            # Check if an explicit persister were given to the call or
            # use the thread-assigned persister otherwise.
            if 'persister' not in kwrds or kwrds['persister'] is None:
                kwrds['persister'] = mcs.thread_local.persister
            return original(*args, **kwrds)
        return _wrap


    def __new__(mcs, cname, cbases, cdict):
        """Create a new class from it's pieces.

        The method will construct a new class by wrapping each method
        of the class using a wrapper that adds the "persistor"
        attribute to the keyword arguments.

        Only callable attributes, class methods, static methods, and
        properties not starting with underscore will be wrapped. All
        other functions are left alone.

        """

        mcs.thread_local.persister = None

        for name, func in cdict.items():
            # Check for functions that should not be touched.

            if name.startswith('_'):
                continue        # Special or internal function

            # Wrap function differently depending on what it is.
            if callable(func):
                # Anything callable is wrapped.
                cdict[name] = mcs.wrapfunc(func)
            elif isinstance(func, (staticmethod, classmethod)):
                # Wrap the inner function of static and class methods
                # and re-wrap using the type constructor.
                cdict[name] = type(func)(mcs.wrapfunc(func.__get__(True)))
            elif isinstance(func, property) and func.fset is not None:
                # Properties are re-constructed from its parts and
                # wrap the fset, fget, and fdel functions.
                newfset = func.fset and mcs.wrapfunc(func.fset)
                newfget = func.fget and mcs.wrapfunc(func.fget)
                newfdel = func.fdel and mcs.wrapfunc(func.fdel)
                cdict[name] = property(newfget, newfset, newfdel, func.__doc__)
        return type.__new__(mcs, cname, cbases, cdict)

    def __init__(cls, cname, cbases, cdict):
        """Add the class to the list of classes that should be called
        on init.
        """
        type.__init__(cls, cname, cbases, cdict)
        PersistentMeta.classes.append(cls)

class Persistable(object):
    """Class for all persistable objects.

    Classes whose instances are going to be persisted to the backing store
    should inherit from this class.

    To be usable by the Fabric the persistence class have to define the
    following methods:

    **create**\ (*persister=None*)
      This is a static (or class) method called when the persistence tables need
      to be created because the system is set up initially. The method is
      expected to use the passed persister to create any tables that are
      necessary to persist the instances of this class.

    **drop**\ (*persister=None*)
      This is a static (or class) method called when the persistence tables need
      to be removed because the system is torn down. The method is expected to
      use the passed persister to drop any tables that where created in the
      `create` method.

    **add_constraint**\ (*persister=None*)
      This is a static (or class) method called after all the persistence tables
      are created in the initial set up. The method is expected to use the
      passed persister to add any constraints needed on the tables (for example,
      foreign keys) created in the **create** method for the class.

    **drop_constraint**\ (*persister=None*)
      This is a static (or class) method called before the persistence tables
      are dropped in the tear down. The method is expected to use the passed
      persister to remove any constraints added to the tables so that the table
      can be trivially dropped later.

    The following methods are customary used to manipulate instances of objects
    and can be defined (but do not have to):

    **add**\ (*object*, *persister=None*)
      This is a static (or class) method to add an instance to the collection of
      instances in the persistent store. This method is expected to use the
      passed persister to insert a new object into the tables for the
      persistable object.

    **remove**\ (*object*, *persister=None*)
      This is a static (or class) method to remove an instance from the
      collection of instances in the persistent store. This method is expected
      to use the passed persister to delete the entry for the object from the
      tables in the persistent store.

    **fetch**\ (..., *persister=None*)
      This is a static (or class) method to fetch an already created instance
      from the collection of instances in the persistent store. This method is
      expected to take any arguments necessary to identify the object and fetch
      the object from the persistent store and return it as a new instance of
      the class.

    .. note::

       All the methods accept a *persister* parameter that should be placed last
       and have a default value of `None`.

    A example for how to create a persistable object is::

        from mysql.fabric.persistence import Persistable

        class Car(Persistable):
            def __init__(self, reg_no, model):
                self.reg_no = reg_no
                self.model = model

            @staticmethod
            def create(persister=None):
                persister.exec_stmt(
                    "CREATE TABLE cars ("
                    "  reg_no CHAR(6) PRIMARY KEY,"
                    "  model VARCHAR(16)"
                    ")"
                )

            @staticmethod
            def drop(persister=None):
                persister.exec_stmt("DROP TABLE cars")

            @staticmethod
            def add(car, persister=None):
                persister.exec_stmt(
                    "INSERT INTO cars(reg_no, model) VALUES (%s)",
                    { "params": (car.reg_no, car.model) }
                )

            @staticmethod
            def remove(car, persister=None):
                persister.exec_stmt(
                    "DELETE FROM cars WHERE reg_no = %s",
                    { "params": (car.reg_no,) }
                )

            @staticmethod
            def fetch(reg_no, persister=None):
                row = persister.exec_stmt(
                    "SELECT reg_no, model FROM cars WHERE reg_no = %s",
                    { "params": (reg_no,) }
                ).fetch_one()

                return Car(reg_no=row[0], model=row[1]) if row else None

    The `add`, `remove`, and `fetch` methods can be used to manipuate instances
    of the peristent class, but in this case the *persister* parameter should
    not be used.  The :class:`PersistentMeta` class will add a wrapper to each
    function that adds the persister argument to the class if it was not
    passed.

    The following is an example of code for manipulating persistent objects::

       my_car = Car(reg_no='XYZ123', model='Volvo')
       Car.add(my_car)
       ...
       some_car = Car.fetch(reg_no='XYZ123')
       print some_car.model
       Car.remove(some_car)

    If the connection to the persistent store is temporarily lost because, for
    example, it has been closed by the server after a period of inactivite, the
    persister will try to automatically restablish it before a new transaction
    is implicitly or explicitly started. Failures while executing a statement
    within a transaction context are automatically reported to the caller.

    The autocommit mode is enabled by default so that a new statement is always
    executed within a new transaction context. If users want to explicitly
    create a transaction context, they can do so as follows::

       import mysql.fabric.persistence

       persister = mysql.fabric.persistence.current_persister()

       persister.begin()
       ...
       persister.commit()
    """

    __metaclass__ = PersistentMeta

class MySQLPersister(object):
    """Class responsible for persisting objects to a MySQL database.

    The class is responsible for managing the connection with the database where
    the persistent objects are stored. There should normally be one instance of
    this class available for each thread.

    Before using the persister, the system has to be initialized using
    :meth:`MySQLPersister.init`. This will provide connection information to the
    object database for each :class:`MySQLPersister` instance created, set up
    the object database, and give each subclass of :class:`Persistable` a chance
    to set itself up by calling the class `init` method.
    """
    @classmethod
    def init(cls, host, user, password=None, port=None, database=None,
             connection_timeout=None, connection_attempts=None,
             connection_delay=None, auth_plugin=None):
        """Initialize the object persistance system.

        This function initializes the persistance system. The function
        is idempotent in the sense that it can be called multiple
        times to set up the system. The parameters provided are passed
        to the connector, so typically, these fields need to be
        present.

        :param host: Host name of the database server.
        :param port: Port of the database server. Defaults to 3306.
        :param user: User to use connecting to the database server.
        :param password: Password to use when connecting to the
                         database server.
        :param database: Database where the persistance information is
                         stored. Default is :const:`DEFAULT_DATABASE`.
        :param connection_timeout: Timeout to connect to the database server.
        :param connection_attempts: Number of attempts to connect or reconnect
                                    to the database server. Default is
                                    :const:`DEFAULT_CONNECT_ATTEMPTS`.
        :param connection_delay: Delay after an atempt to connect or reconnect
                                 to the database server. Default is
                                 :const:`DEFAULT_CONNECT_DELAY`.
        :param auth_plugin: Use auth_plugin as authencation plugin for
                            authentication with the database server.
        """
        if port is None:
            port = _server_utils.MYSQL_DEFAULT_PORT

        if database is None:
            database = DEFAULT_DATABASE

        if connection_attempts is None:
            connection_attempts = DEFAULT_CONNECT_ATTEMPTS

        if connection_delay is None:
            connection_delay = DEFAULT_CONNECT_DELAY

        # Save away the connection information, it will be used by the
        # threads.
        cls.connection_info = {
            "host": host, "port": port,
            "user": user, "password": password,
            "connection_timeout" : connection_timeout,
            "auth_plugin": auth_plugin,
            }
        cls.connection_attempts = connection_attempts
        cls.connection_delay = connection_delay
        cls.database = database

    @classmethod
    def setup(cls):
        """Setup the object persistance system.

        Perform initialization, which in this case means creating
        the database if it does not exist.
        """
        assert (cls.connection_info is not None)
        conn = _server_utils.create_mysql_connection(
            autocommit=True, use_unicode=False, **cls.connection_info
        )
        _server_utils.exec_mysql_stmt(
            conn, "CREATE DATABASE %s" % (cls.database, )
        )

    @classmethod
    def teardown(cls):
        """Tear down the object persistance system.

        This should only be called if the persistance database should
        be removed from the persistance server since it will delete
        all object tables.
        """
        assert (cls.connection_info is not None)
        conn = _server_utils.create_mysql_connection(
            autocommit=True, use_unicode=False, **cls.connection_info
        )
        _server_utils.exec_mysql_stmt(
            conn, "DROP DATABASE IF EXISTS %s" % (cls.database, )
        )

    def __init__(self):
        """Constructor for MySQLPersister.
        """
        self.__cnx = None
        self.__check_connection = True

        assert (self.connection_info is not None)
        try:
            self.__cnx = _server_utils.create_mysql_connection(
                autocommit=True, use_unicode=False, database=self.database,
                **self.connection_info
            )
        except _errors.DatabaseError:
            pass

        if self.uuid is None and self.__cnx is not None:
            _LOGGER.warning(
                "Backing store does not support UUID (or not configured "
                "with UUID)."
            )

    def __del__(self):
        """Destructor for MySQLPersister.
        """
        try:
            if self.__cnx:
                _server_utils.destroy_mysql_connection(self.__cnx)
        except AttributeError:
            pass

    def begin(self):
        """Start a new transaction.
        """
        self.exec_stmt("BEGIN")
        self.__check_connection = False

    def commit(self):
        """Commit an on-going transaction.
        """
        try:
            self.exec_stmt("COMMIT")
        finally:
            self.__check_connection = True

    def rollback(self):
        """Roll back an on-going transaction.
        """
        try:
            self.exec_stmt("ROLLBACK")
        finally:
            self.__check_connection = True

    def auth_mysql_token(self):
        """Returns the authentication plugin data found in handshake"""
        return self.__cnx._handshake['scramble']

    @property
    def uuid(self):
        """Return the MySQLPersister's UUID if the server supports it.
        Otherwise, return None.
        """
        try:
            row = _server_utils.exec_mysql_stmt(
                self.__cnx, "SELECT @@GLOBAL.SERVER_UUID"
            )
            return _uuid.UUID(row[0][0])
        except _errors.DatabaseError:
            pass

        return None

    def max_allowed_connections(self):
        """Return the maximum number of allowed connections to server.
        """
        row = _server_utils.exec_mysql_stmt(
            self.__cnx, "SELECT @@GLOBAL.max_connections"
        )
        return int(row[0][0])

    def exec_stmt(self, stmt_str, options=None):
        """Execute statements against the server.

        If a new transaction is about to be started, this method checks whether
        the connection is valid or not. If the connection is invalid, it tries
        to restablish it as MySQL might disconnect inactive connections.

        See :meth:`~mysql.fabric.server_utils.exec_stmt`.
        """
        while True:
            if self.__check_connection and \
                not _server_utils.is_valid_mysql_connection(self.__cnx):
                self._try_to_fix_connection()
            return _server_utils.exec_mysql_stmt(
                self.__cnx, stmt_str, options
            )

    def _try_to_fix_connection(self):
        """Try to get a new connection if the current one is stale.
        """
        for attempt in range(0, self.connection_attempts):
            try:
                if self.__cnx:
                    _server_utils.reestablish_mysql_connection(
                        self.__cnx, attempt=1, delay=0
                    )
                else:
                    self.__cnx = _server_utils.create_mysql_connection(
                        autocommit=True, use_unicode=False,
                        database=self.database, **self.connection_info
                    )
                return
            except _errors.DatabaseError as error:
                _LOGGER.debug("Error accessing backing store (%s). "
                    "Attempt (%s).", error, attempt)
            time.sleep(self.connection_delay)

def current_persister():
    """Return the persister for the current thread.
    """
    return PersistentMeta.thread_local.persister

def init_thread():
    """Initialize the persistence system for the thread.
    """
    PersistentMeta.init_thread(MySQLPersister())

def deinit_thread():
    """Initialize the persistence system for the thread.
    """
    PersistentMeta.deinit_thread()

_LOGGER = logging.getLogger(__name__)

def init(host, user, password=None, port=None, database=None,
         connection_timeout=None, connection_attempts=None,
         connection_delay=None, auth_plugin=None):
    """Initialize the persistance system.

    This function is idempotent in the sense that it can be executed
    multiple times without destroying anything. This property is
    important since normally the :func:`setup` and :func:`teardown`
    are not idempotent.

    :param host: Hostname to connect to.
    :param user: User to connect as.
    :param password: Password to use when connecting. Default to the
                     empty password.
    :param port: Port to connect to. Default to 3306.
    :param database: Database to store object data in. Default to
                     :const:`DEFAULT_DATABASE`.
    :param connection_timeout: Timeout to connect to the database server.
    :param connection_attempts: Number of attempts to connect or reconnect to
                                the database server. Default is
                                :const:`DEFAULT_CONNECT_ATTEMPTS`.
    :param connection_delay: Delay after an atempt to connect or reconnect
                             to the database server. Default is
                             :const:`DEFAULT_CONNECT_DELAY`.
    """
    _LOGGER.info(
        "Initializing persister: user (%s), server (%s:%d), database (%s).",
        user, host, port, database
    )

    MySQLPersister.init(
        host=host, port=port, user=user, password=password, database=database,
        connection_timeout=connection_timeout,
        connection_attempts=connection_attempts,
        connection_delay=connection_delay,
        auth_plugin=auth_plugin
    )

def setup(config=None):
    """ Setup the persistance system globally.

    This means creating any databases, tables and constraints necessary in the
    persistence database.
    """
    MySQLPersister.setup()

    persister = MySQLPersister()
    try:
        for cls in PersistentMeta.classes:
            if hasattr(cls, 'create'):
                _LOGGER.debug("Create database objects for %s", cls.__name__)
                if 'config' in inspect.getargspec(cls.create):
                    cls.create(persister=persister, config=config)
                else:
                    # create() does not support config
                    cls.create(persister=persister)

        #Initialize the constraints after creating the tables.
        for cls in PersistentMeta.classes:
            #Call the add_constraints method of those classes that sub-class from
            #Persistence and those which have an implementation of add_constraints.
            if hasattr(cls, 'add_constraints'):
                _LOGGER.debug("Create constraints for %s", cls.__name__)
                cls.add_constraints(persister=persister)
    except _errors.DatabaseError:
        MySQLPersister.teardown()
        raise

def teardown():
    """Teardown the persistance system globally.

    This means removing any tables,constraints created. Normally, this
    function does not have to be executed on shutdown since that would
    emove all necessary tables.
    """
    _LOGGER.info("Teardown persister.")
    MySQLPersister.teardown()



#
# Copyright (c) 2010, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
This module contains methods for working with mysql server tools.
"""

import inspect
import os
import re
import sys
import shutil
import socket
import subprocess
import time

from mysql.utilities import (PYTHON_MIN_VERSION, PYTHON_MAX_VERSION,
                             CONNECTOR_MIN_VERSION)
from mysql.utilities.common.format import print_list
from mysql.utilities.exception import UtilError


def _add_basedir(search_paths, path_str):
    """Add a basedir and all known sub directories

    This method builds a list of possible paths for a basedir for locating
    special MySQL files like mysqld (mysqld.exe), etc.

    search_paths[inout] List of paths to append
    path_str[in]        The basedir path to append
    """
    search_paths.append(path_str)
    search_paths.append(os.path.join(path_str, "sql"))       # for source trees
    search_paths.append(os.path.join(path_str, "client"))    # for source trees
    search_paths.append(os.path.join(path_str, "share"))
    search_paths.append(os.path.join(path_str, "scripts"))
    search_paths.append(os.path.join(path_str, "bin"))
    search_paths.append(os.path.join(path_str, "libexec"))
    search_paths.append(os.path.join(path_str, "mysql"))


def get_tool_path(basedir, tool, fix_ext=True, required=True,
                  defaults_paths=None, search_PATH=False):
    """Search for a MySQL tool and return the full path

    basedir[in]         The initial basedir to search (from mysql server)
    tool[in]            The name of the tool to find
    fix_ext[in]         If True (default is True), add .exe if running on
                        Windows.
    required[in]        If True (default is True), and error will be
                        generated and the utility aborted if the tool is
                        not found.
    defaults_paths[in]  Default list of paths to search for the tool.
                        By default an empty list is assumed, i.e. [].
    search_PATH[in]     Boolean value that indicates if the paths specified by
                        the PATH environment variable will be used to search
                        for the tool. By default the PATH will not be searched,
                        i.e. search_PATH=False.
    Returns (string) full path to tool
    """
    if not defaults_paths:
        defaults_paths = []
    search_paths = []

    if basedir:
        # Add specified basedir path to search paths
        _add_basedir(search_paths, basedir)
    if defaults_paths and len(defaults_paths):
        # Add specified default paths to search paths
        for path in defaults_paths:
            search_paths.append(path)
    else:
        # Add default basedir paths to search paths
        _add_basedir(search_paths, "/usr/local/mysql/")
        _add_basedir(search_paths, "/usr/sbin/")
        _add_basedir(search_paths, "/usr/share/")

    # Search in path from the PATH environment variable
    if search_PATH:
        for path in os.environ['PATH'].split(os.pathsep):
            search_paths.append(path)

    if os.name == "nt" and fix_ext:
        tool = tool + ".exe"
    # Search for the tool
    for path in search_paths:
        norm_path = os.path.normpath(path)
        if os.path.isdir(norm_path):
            toolpath = os.path.join(norm_path, tool)
            if os.path.isfile(toolpath):
                return toolpath
            else:
                if tool == "mysqld.exe":
                    toolpath = os.path.join(norm_path, "mysqld-nt.exe")
                    if os.path.isfile(toolpath):
                        return toolpath
    if required:
        raise UtilError("Cannot find location of %s." % tool)

    return None


def delete_directory(path):
    """Remove a directory (folder) and its contents.

    path[in]           target directory
    """
    if os.path.exists(path):
        # It can take up to 10 seconds for Windows to 'release' a directory
        # once a process has terminated. We wait...
        if os.name == "nt":
            stop = 10
            i = 1
            while i < stop and os.path.exists(path):
                shutil.rmtree(path, True)
                time.sleep(1)
                i += 1
        else:
            shutil.rmtree(path, True)


def execute_script(run_cmd, filename=None, options=None, verbosity=False):
    """Execute a script.

    This method spawns a subprocess to execute a script. If a file is
    specified, it will direct output to that file else it will suppress
    all output from the script.

    run_cmd[in]        command/script to execute
    filename[in]       file path name to file, os.stdout, etc.
                       Default is None (do not log/write output)
    options[in]        arguments for script
                       Default is no arguments ([])
    verbosity[in]      show result of script
                       Default is False

    Returns int - result from process execution
    """
    if options is None:
        options = []
    if verbosity:
        f_out = sys.stdout
    else:
        if not filename:
            filename = os.devnull
        f_out = open(filename, 'w')

    str_opts = [str(opt) for opt in options]
    cmd_opts = " ".join(str_opts)
    command = " ".join([run_cmd, cmd_opts])

    if verbosity:
        print("# SCRIPT EXECUTED: {0}".format(command))

    proc = subprocess.Popen(command, shell=True,
                            stdout=f_out, stderr=f_out)
    ret_val = proc.wait()
    if not verbosity:
        f_out.close()
    return ret_val


def ping_host(host, timeout):
    """Execute 'ping' against host to see if it is alive.

    host[in]           hostname or IP to ping
    timeout[in]        timeout in seconds to wait

    returns bool - True = host is reachable via ping
    """
    if sys.platform == "darwin":
        run_cmd = "ping -o -t %s %s" % (timeout, host)
    elif os.name == "posix":
        run_cmd = "ping -w %s %s" % (timeout, host)
    else:  # must be windows
        run_cmd = "ping -n %s %s" % (timeout, host)

    ret_val = execute_script(run_cmd)

    return (ret_val == 0)


def get_mysqld_version(mysqld_path):
    """Return the version number for a mysqld executable.

    mysqld_path[in]    location of the mysqld executable

    Returns tuple - (major, minor, release), or None if error
    """
    out = open("version_check", 'w')
    proc = subprocess.Popen("%s --version" % mysqld_path,
                            stdout=out, stderr=out, shell=True)
    proc.wait()
    out.close()
    out = open("version_check", 'r')
    line = None
    for line in out.readlines():
        if "Ver" in line:
            break
    out.close()

    try:
        os.unlink('version_check')
    except:
        pass

    if line is None:
        return None
    version = line.split(' ', 5)[3]
    try:
        maj_ver, min_ver, dev = version.split(".")
        rel = dev.split("-")
        return (maj_ver, min_ver, rel[0])
    except:
        return None

    return None


def show_file_statistics(file_name, wild=False, out_format="GRID"):
    """Show file statistics for file name specified

    file_name[in]    target file name and path
    wild[in]         if True, get file statistics for all files with prefix of
                     file_name. Default is False
    out_format[in]   output format to print file statistics. Default is GRID.
    """

    def _get_file_stats(path, file_name):
        """Return file stats
        """
        stats = os.stat(os.path.join(path, file_name))
        return ((file_name, stats.st_size, time.ctime(stats.st_ctime),
                 time.ctime(stats.st_mtime)))

    columns = ["File", "Size", "Created", "Last Modified"]
    rows = []
    path, filename = os.path.split(file_name)
    if wild:
        for _, _, files in os.walk(path):
            for f in files:
                if f.startswith(filename):
                    rows.append(_get_file_stats(path, f))
    else:
        rows.append(_get_file_stats(path, filename))

    print_list(sys.stdout, out_format, columns, rows)


def remote_copy(filepath, user, host, local_path, verbosity=0):
    """Copy a file from a remote machine to the localhost.

    filepath[in]       The full path and file name of the file on the remote
                       machine
    user[in]           Remote login
    local_path[in]     The path to where the file is to be copie

    Returns bool - True = succes, False = failure or exception
    """

    if os.name == "posix":  # use scp
        run_cmd = "scp %s@%s:%s %s" % (user, host, filepath, local_path)
        if verbosity > 1:
            print("# Command =%s" % run_cmd)
        print("# Copying file from %s:%s to %s:" %
              (host, filepath, local_path))
        proc = subprocess.Popen(run_cmd, shell=True)
        proc.wait()
    else:
        print("Remote copy not supported. Please use UNC paths and omit "
              "the --remote-login option to use a local copy operation.")
    return True


def check_python_version(min_version=PYTHON_MIN_VERSION,
                         max_version=PYTHON_MAX_VERSION,
                         raise_exception_on_fail=False,
                         name=None, print_on_fail=True,
                         exit_on_fail=True,
                         return_error_msg=False):
    """Check the Python version compatibility.

    By default this method uses constants to define the minimum and maximum
    Python versions required. It's possible to override this by passing new
    values on ``min_version`` and ``max_version`` parameters.
    It will run a ``sys.exit`` or raise a ``UtilError`` if the version of
    Python detected it not compatible.

    min_version[in]               Tuple with the minimum Python version
                                  required (inclusive).
    max_version[in]               Tuple with the maximum Python version
                                  required (exclusive).
    raise_exception_on_fail[in]   Boolean, it will raise a ``UtilError`` if
                                  True and Python detected is not compatible.
    name[in]                      String for a custom name, if not provided
                                  will get the module name from where this
                                  function was called.
    print_on_fail[in]             If True, print error else do not print
                                  error on failure.
    exit_on_fail[in]              If True, issue exit() else do not exit()
                                  on failure.
    return_error_msg[in]          If True, and is not compatible
                                  returns (result, error_msg) tuple.
    """

    # Only use the fields: major, minor and micro
    sys_version = sys.version_info[:3]

    # Test min version compatibility
    is_compat = min_version <= sys_version

    # Test max version compatibility if it's defined
    if is_compat and max_version:
        is_compat = sys_version < max_version

    if not is_compat:
        if not name:
            # Get the utility name by finding the module
            # name from where this function was called
            frm = inspect.stack()[1]
            mod = inspect.getmodule(frm[0])
            mod_name = os.path.splitext(
                os.path.basename(mod.__file__))[0]
            name = '{0} utility'.format(mod_name)

        # Build the error message
        if max_version:
            max_version_error_msg = 'or higher and lower than %s' % \
                '.'.join([str(el) for el in max_version])
        else:
            max_version_error_msg = 'or higher'

        error_msg = (
            'The %(name)s requires Python version %(min_version)s '
            '%(max_version_error_msg)s. The version of Python detected was '
            '%(sys_version)s. You may need to install or redirect the '
            'execution of this utility to an environment that includes a '
            'compatible Python version.'
        ) % {
            'name': name,
            'sys_version': '.'.join([str(el) for el in sys_version]),
            'min_version': '.'.join([str(el) for el in min_version]),
            'max_version_error_msg': max_version_error_msg
        }

        if raise_exception_on_fail:
            raise UtilError(error_msg)

        if print_on_fail:
            print('ERROR: {0}'.format(error_msg))

        if exit_on_fail:
            sys.exit(1)

        if return_error_msg:
            return is_compat, error_msg

    return is_compat


def check_port_in_use(host, port):
    """Check to see if port is in use.

    host[in]            Hostname or IP to check
    port[in]            Port number to check

    Returns bool - True = port is available, False is not available
    """
    try:
        sock = socket.create_connection((host, port))
    except socket.error:
        return True
    sock.close()
    return False


def requires_encoding(orig_str):
    r"""Check to see if a string requires encoding

    This method will check to see if a string requires encoding to be used
    as a MySQL file name (r"[\w$]*").

    orig_str[in]        original string

    Returns bool - True = requires encoding, False = does not require encoding
    """
    ok_chars = re.compile(r"[\w$]*")
    parts = ok_chars.findall(orig_str)
    return len(parts) > 2 and parts[1].strip() == ''


def encode(orig_str):
    r"""Encode a string containing non-MySQL observed characters

    This method will take a string containing characters other than those
    recognized by MySQL (r"[\w$]*") and covert them to embedded ascii values.
    For example, "this.has.periods" becomes "this@002ehas@00e2periods"

    orig_str[in]        original string

    Returns string - encoded string or original string
    """
    # First, find the parts that match valid characters
    ok_chars = re.compile(r"[\w$]*")
    parts = ok_chars.findall(orig_str)

    # Now find each part that does not match the list of valid characters
    # Save the good parts
    i = 0
    encode_parts = []
    good_parts = []
    for part in parts:
        if not len(part):
            continue
        good_parts.append(part)
        if i == 0:
            i = len(part)
        else:
            j = orig_str[i:].find(part)
            encode_parts.append(orig_str[i:i + j])
            i += len(part) + j

    # Next, convert the non-valid parts to the form @NNNN (hex)
    encoded_parts = []
    for part in encode_parts:
        new_part = "".join(["@%04x" % ord(c) for c in part])
        encoded_parts.append(new_part)

    # Take the good parts and the encoded parts and reform the string
    i = 0
    new_parts = []
    for part in good_parts[:len(good_parts) - 1]:
        new_parts.append(part)
        new_parts.append(encoded_parts[i])
        i += 1
    new_parts.append(good_parts[len(good_parts) - 1])

    # Return the new string
    return "".join(new_parts)


def requires_decoding(orig_str):
    """Check to if a string required decoding

    This method will check to see if a string requires decoding to be used
    as a filename (has @NNNN entries)

    orig_str[in]        original string

    Returns bool - True = requires decoding, False = does not require decoding
    """
    return '@' in orig_str


def decode(orig_str):
    r"""Decode a string containing @NNNN entries

    This method will take a string containing characters other than those
    recognized by MySQL (r"[\w$]*") and covert them to character values.
    For example, "this@002ehas@00e2periods" becomes "this.has.periods".

    orig_str[in]        original string

    Returns string - decoded string or original string
    """
    parts = orig_str.split('@')
    if len(parts) == 1:
        return orig_str
    new_parts = [parts[0]]
    for part in parts[1:]:
        # take first four positions and convert to ascii
        new_parts.append(chr(int(part[0:4], 16)))
        new_parts.append(part[4:])
    return "".join(new_parts)


def check_connector_python(print_error=True,
                           min_version=CONNECTOR_MIN_VERSION):

    """Check to see if Connector Python is installed and accessible and
    meets minimum required version.

    By default this method uses constants to define the minimum
    C/Python version required. It's possible to override this by passing  a new
    value to ``min_version`` parameter.

    print_error[in]               If True, print error else do not print
                                  error on failure.
    min_version[in]               Tuple with the minimum C/Python version
                                  required (inclusive).

    """
    is_compatible = True
    try:
        import mysql.connector  # pylint: disable=W0612
    except ImportError:
        if print_error:
            print("ERROR: The MySQL Connector/Python module was not found. "
                  "MySQL Utilities requires the connector to be installed. "
                  "Please check your paths or download and install the "
                  "Connector/Python from http://dev.mysql.com.")
        return False
    else:
        try:
            sys_version = mysql.connector.version.VERSION[:3]
        except AttributeError:
            is_compatible = False

    if is_compatible and sys_version >= min_version:
        return True
    else:
        if print_error:
            print("ERROR: The MYSQL Connector/Python module was found "
                  "but it is either not properly installed or it is an "
                  "old version. MySQL Utilities requires Connector/Python "
                  "version > '{0}'. Download and install Connector/Python "
                  "from http://dev.mysql.com.".format(min_version))
        return False


def print_elapsed_time(start_time):
    """Print the elapsed time to stdout (screen)

    start_time[in]      The starting time of the test
    """
    stop_time = time.time()
    display_time = stop_time - start_time
    print("Time: {0:.2f} sec\n".format(display_time))



#
# Copyright (c) 2014 Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#
"""Provides an abstraction that represents the Fabric Node.
"""
import uuid as _uuid
from mysql.fabric.utils import Singleton

class FabricNode(Singleton):
    """An abstraction that represents the Fabric Node.

    :param uuid: The UUID for the FabricNode. None if a UUID should
                 be automatically generated.
    """
    def __init__(self, uuid=None):
        """Constructor for FabricNode.
        """
        assert(uuid is None or isinstance(uuid, _uuid.UUID))
        # This is temporary set to 0 to make it compatible with FABRIC_UUID in
        # utils.py.
        self.__uuid = 0
        self.startup = None

    @property
    def uuid(self):
        """Return Fabric identification which represented as a UUID.
        """
        return self.__uuid



#
# Copyright (c) 2013,2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
#
"""This module provides the necessary interfaces for performing administrative
tasks on groups and servers, specifically MySQL Servers.

It is possible to add, update and remove a group. One cannot, however, remove
a group if there are associated servers. It is possible to add a server to a
group and remove a server from a group. Search functions are also provided so
that one may look up groups and servers. Given a server's address, one may
also find out the server's uuid if the server is alive and kicking.

When a group is created though, it is inactive which means that the failure
detector will not check if its servers are alive. To start up the failure
detector, one needs to explicitly activate it per group. A server may have
one of the following statuses:

- PRIMARY - This is set when the server may accept both reads and writes
  operations.
- SECONDARY - This is set when the server may accept only read operations.
- SPARE - This is set when users want to have server that is kept in sync
  but does not accept neither reads or writes operations.
- FAULTY - This is set by the failure detector and indicates that a server
  is not reachable.

Find in what follows the possible state transitions:

.. graphviz::

   digraph state_transition {
    rankdir=LR;
    size="8,5"

    node [shape = circle]; Primary;
    node [shape = circle]; Secondary;
    node [shape = circle]; Spare;
    node [shape = circle]; Faulty;

    Primary   -> Secondary [ label = "demote" ];
    Primary   -> Faulty [ label = "failure" ];
    Secondary -> Primary [ label = "promote" ];
    Secondary -> Spare [ label = "set_status" ];
    Secondary -> Faulty [ label = "failure" ];
    Spare     -> Primary [ label = "promote" ];
    Spare     -> Secondary [ label = "set_status" ];
    Spare     -> Faulty [ label = "failure" ];
    Faulty    -> Spare [ label = "set_status" ];
  }

It is worth noticing that this module only provides functions for performing
basic administrative tasks, provisioning and high-availability functions are
provided elsewhere.
"""
import logging
import uuid as _uuid

import mysql.fabric.services.utils as _utils

from mysql.fabric import (
    backup as _backup,
    events as _events,
    server as _server,
    errors as _errors,
    failure_detector as _detector,
    group_replication as _group_replication,
    sharding as _sharding,
    server_utils as _server_utils,
)

from mysql.fabric.command import (
    ProcedureGroup,
    Command,
)

from mysql.fabric.utils import (
    get_time,
)

_LOGGER = logging.getLogger(__name__)

SERVER_NOT_FOUND = "Server with UUID %s not found."
MYSQLDUMP_NOT_FOUND = "Unable to find MySQLDump in location %s"
MYSQLCLIENT_NOT_FOUND = "Unable to find MySQL Client in location %s"

class GroupLookups(Command):
    """Return information on existing group(s).
    """
    group_name = "group"
    command_name = "lookup_groups"

    def execute(self, group_id=None):
        """Return information on existing group(s).

        :param group_id: None if one wants to list the existing groups or
                         group's id if one wants information on a group.
        :return: List with {"group_id" : group_id, "failure_detector": ON/OFF,
                 "description" : description}.
        """
        return Command.generate_output_pattern(
            _lookup_groups, group_id)

CREATE_GROUP = _events.Event()
class GroupCreate(ProcedureGroup):
    """Create a group.
    """
    group_name = "group"
    command_name = "create"

    def execute(self, group_id, description=None, synchronous=True):
        """Create a group.

        :param group_id: Group's id.
        :param description: Group's description.
        :param synchronous: Whether one should wait until the execution finishes
                            or not.
        :return: Tuple with job's uuid and status.
        """
        procedures = _events.trigger(
            CREATE_GROUP, self.get_lockable_objects(), group_id, description
        )
        return self.wait_for_procedures(procedures, synchronous)

UPDATE_GROUP = _events.Event()
class GroupDescription(ProcedureGroup):
    """Update group's description.
    """
    group_name = "group"
    command_name = "description"

    def execute(self, group_id, description=None, synchronous=True):
        """Update group's description.

        :param group_id: Group's id.
        :param description: Group's description.
        :param synchronous: Whether one should wait until the execution finishes
                            or not.
        :return: Tuple with job's uuid and status.
        """
        procedures = _events.trigger(
            UPDATE_GROUP, self.get_lockable_objects(), group_id, description
        )
        return self.wait_for_procedures(procedures, synchronous)

DESTROY_GROUP = _events.Event()
class DestroyGroup(ProcedureGroup):
    """Remove a group.
    """
    group_name = "group"
    command_name = "destroy"

    def execute(self, group_id, force=False, synchronous=True):
        """Remove a group.

        :param group_id: Group's id.
        :param force: If the group is not empty, remove it serves.
        :param synchronous: Whether one should wait until the execution finishes
                            or not.
        :return: Tuple with job's uuid and status.
        """
        procedures = _events.trigger(
            DESTROY_GROUP, self.get_lockable_objects(), group_id, force
        )
        return self.wait_for_procedures(procedures, synchronous)

class ServerLookups(Command):
    """Return information on existing server(s) in a group.
    """
    group_name = "group"
    command_name = "lookup_servers"

    def execute(self, group_id, server_id=None, status=None, mode=None):
        """Return information on existing server(s) in a group.

        :param group_id: Group's id.
        :param uuid: None if one wants to list the existing servers
                     in a group or server's id if one wants information
                     on a server in a group.
        :server_id type: Servers's UUID or HOST:PORT.
        :param status: Server's status one is searching for.
        :param mode: Server's mode one is searching for.
        :return: Information on servers.
        :rtype: List with [uuid, address, status, mode, weight]
        """
        return Command.generate_output_pattern(
            _lookup_servers, group_id, server_id, status, mode
        )

class ServerUuid(Command):
    """Return server's uuid.
    """
    group_name = "server"
    command_name = "lookup_uuid"

    def execute(self, address, timeout=5):
        """Return server's UUID.

        :param address: Server's address.
        :param timeout: Time in seconds after which an error is reported
                        if the UUID is not retrieved.
        :return: UUID.
        """
        return Command.generate_output_pattern(_lookup_uuid, address, timeout)

ADD_SERVER = _events.Event()
class ServerAdd(ProcedureGroup):
    """Add a server into group.

    If users just want to update the state store and skip provisioning steps
    such as configuring replication, the update_only parameter must be set to
    true.

    Note that the current implementation has a simple provisioning step that
    makes the server point to the master if there is any.
    """
    group_name = "group"
    command_name = "add"

    def execute(self, group_id, address, timeout=5, update_only=False,
                synchronous=True):
        """Add a server into a group.

        :param group_id: Group's id.
        :param address: Server's address.
        :param timeout: Time in seconds after which an error is reported
                        if one cannot access the server.
        :update_only: Only update the state store and skip provisioning.
        :param synchronous: Whether one should wait until the execution
                            finishes or not.
        :return: Tuple with job's uuid and status.
        """
        procedures = _events.trigger(ADD_SERVER, self.get_lockable_objects(),
            group_id, address, timeout, update_only
        )
        return self.wait_for_procedures(procedures, synchronous)

REMOVE_SERVER = _events.Event()
class ServerRemove(ProcedureGroup):
    """Remove a server from a group.
    """
    group_name = "group"
    command_name = "remove"

    def execute(self, group_id, server_id, synchronous=True):
        """Remove a server from a group.

        :param group_id: Group's id.
        :param server_id: Servers's UUID or HOST:PORT.
        :param synchronous: Whether one should wait until the execution
                            finishes or not.
        :return: Tuple with job's uuid and status.
        """
        procedures = _events.trigger(REMOVE_SERVER, self.get_lockable_objects(),
            group_id, server_id
        )
        return self.wait_for_procedures(procedures, synchronous)

ACTIVATE_GROUP = _events.Event()
class ActivateGroup(ProcedureGroup):
    """Activate failure detector for server(s) in a group.

    By default the failure detector is disabled.
    """
    group_name = "group"
    command_name = "activate"

    def execute(self, group_id, synchronous=True):
        """Activate a group.

        :param group_id: Group's id.
        :param synchronous: Whether one should wait until the execution
                            finishes or not.
        :return: Tuple with job's uuid and status.
        """
        procedures = _events.trigger(
            ACTIVATE_GROUP, self.get_lockable_objects(), group_id
        )
        return self.wait_for_procedures(procedures, synchronous)

DEACTIVATE_GROUP = _events.Event()
class DeactivateGroup(ProcedureGroup):
    """Deactivate failure detector for server(s) in a group.

    By default the failure detector is disabled.
    """
    group_name = "group"
    command_name = "deactivate"

    def execute(self, group_id, synchronous=True):
        """Deactivate group.

        :param group_id: Group's id.
        :param synchronous: Whether one should wait until the execution
                            finishes or not.
        :return: Tuple with job's uuid and status.
        """
        procedures = _events.trigger(
            DEACTIVATE_GROUP, self.get_lockable_objects(), group_id
        )
        return self.wait_for_procedures(procedures, synchronous)

class DumpServers(Command):
    """Return information about servers.

    The servers might belong to any group that matches any of the provided
    patterns, or all servers if no patterns are provided.
    """
    group_name = "dump"
    command_name = "servers"

    def execute(self, connector_version=None, patterns=""):
        """Return information about all servers.

        :param connector_version: The connectors version of the data.
        :param patterns: group pattern.
        """
        return _server.MySQLServer.dump_servers(connector_version, patterns)

SET_SERVER_STATUS = _events.Event()
class SetServerStatus(ProcedureGroup):
    """Set a server's status.

    Any server added into a group has to be alive and kicking and its status
    is automatically set to SECONDARY. If the failure detector is activate
    and the server is not reachable, it is automatically set to FAULTY.

    Users can also manually change the server's status. Usually, a user may
    change a slave's mode to SPARE to avoid write and read access and
    guarantee that it is not choosen when a failover or swithover routine is
    executed.

    By default replication is automatically configured when a server has its
    status changed. In order to skip this, users must set the update_only
    parameter to true. If done so, only the state store will be updated with
    information on the new status.
    """
    group_name = "server"
    command_name = "set_status"

    def execute(self, server_id, status, update_only=False, synchronous=True):
        """Set a server's status.

        :param server_id: Servers's UUID or HOST:PORT.
        :param status: Server's status.
        :update_only: Only update the state store and skip provisioning.
        :param synchronous: Whether one should wait until the execution
                            finishes or not.
        """
        procedures = _events.trigger(
            SET_SERVER_STATUS, self.get_lockable_objects(), server_id, status,
            update_only
        )
        return self.wait_for_procedures(procedures, synchronous)

SET_SERVER_WEIGHT = _events.Event()
class SetServerWeight(ProcedureGroup):
    """Set a server's weight.

    The weight determines the likelihood of a server being choseen by a
    connector to process transactions. For example, a server whose weight
    is 2.0 may receive 2 times more requests than a server whose weight is
    1.0.
    """
    group_name = "server"
    command_name = "set_weight"

    def execute(self, server_id, weight, synchronous=True):
        """Set a server's weight.

        :param server_id: Servers's UUID or HOST:PORT.
        :param weight: Server's weight.
        :param synchronous: Whether one should wait until the execution
                            finishes or not.
        """
        procedures = _events.trigger(
            SET_SERVER_WEIGHT, self.get_lockable_objects(), server_id, weight
        )
        return self.wait_for_procedures(procedures, synchronous)

SET_SERVER_MODE = _events.Event()
class SetServerMode(ProcedureGroup):
    """Set a server's mode.

    The mode determines whether a server can process read-only, read-write
    or both transaction types.
    """
    group_name = "server"
    command_name = "set_mode"

    def execute(self, server_id, mode, synchronous=True):
        """Set a server's mode.

        :param server_id: Servers's UUID or HOST:PORT.
        :param weight: Server's weight.
        :param synchronous: Whether one should wait until the execution
                            finishes or not.
        """
        procedures = _events.trigger(
            SET_SERVER_MODE, self.get_lockable_objects(), server_id, mode
        )
        return self.wait_for_procedures(procedures, synchronous)

BACKUP_SERVER = _events.Event("BACKUP_SERVER")
RESTORE_SERVER = _events.Event("RESTORE_SERVER")
class CloneServer(ProcedureGroup):
    """Clone the objects of a given server into a destination server.
    """
    group_name = "server"
    command_name = "clone"

    def execute(self, group_id, destn_address, server_uuid=None,
                synchronous=True):
        """Clone the objects of a given server into a destination server.

        :param group_id: The ID of the source group.
        :param destn_address: The address of the MySQL Server to which
            the clone needs to happen.
        :param server_uuid: The UUID of the source MySQL Server.
        :param synchronous: Whether one should wait until the execution
                            finishes or not.
        """
        #If the destination server is already part of a Fabric Group, raise
        #an error
        if destn_address:
            destn_server_uuid = _server.MySQLServer.\
                discover_uuid(destn_address)
            destn_server = _server.MySQLServer.fetch(destn_server_uuid)
            #we should check for both the presence of the server object
            #and its associated group ID. Checking its association with
            #a group ID would verify that a server that is part of Fabric
            #but is not part of a group can also be cloned into.
            if destn_server and destn_server.group_id:
                raise _errors.ServerError("The Destination server is already "\
                    "part of Group (%s)" % (destn_server.group_id,))

        config_file = self.config.config_file if self.config.config_file else ""

        mysqldump_binary = _utils.read_config_value(
                                self.config,
                                'sharding',
                                'mysqldump_program'
                            )
        mysqlclient_binary = _utils.read_config_value(
                                self.config,
                                'sharding',
                                'mysqlclient_program'
                            )

        if not _utils.is_valid_binary(mysqldump_binary):
            raise _errors.ServerError(MYSQLDUMP_NOT_FOUND % mysqldump_binary)

        if not _utils.is_valid_binary(mysqlclient_binary):
            raise _errors.ServerError(MYSQLCLIENT_NOT_FOUND % mysqlclient_binary)

        if server_uuid:
            server = _server.MySQLServer.fetch(server_uuid)
            if group_id != server.group_id:
                raise _errors.ServerError("The server %s was not found in "\
                                          "group %s" % (server_uuid, group_id, ))
        elif not server_uuid:
            group = _server.Group.fetch(group_id)
            server = _utils.fetch_backup_server(group)
            server_uuid = str(server.uuid)

        host, port = _server_utils.split_host_port(
                                destn_address,
                                _backup.MySQLDump.MYSQL_DEFAULT_PORT
                            )

        procedures = _events.trigger(
            BACKUP_SERVER,
            self.get_lockable_objects(),
            server_uuid,
            host,
            port,
            mysqldump_binary,
            mysqlclient_binary,
            config_file
        )
        return self.wait_for_procedures(procedures, synchronous)

def _lookup_groups(group_id=None):
    """Return a list of existing group(s).
    """
    info = []
    if group_id is None:
        for group_id in _server.Group.groups():
            _group_information(_server.Group.fetch(group_id[0]), info)
    else:
        group = _retrieve_group(group_id)
        _group_information(group, info)
    return info

def _group_information(group, info):
    """Get information on group and append it into to a list.
    """
    master = str(group.master) if group.master else ""
    info.append({
        "group_id" : group.group_id,
        "description" : group.description or "",
        "failure_detector" : True if group.status else False,
        "master_uuid" : master
    })

@_events.on_event(CREATE_GROUP)
def _create_group(group_id, description):
    """Create a group.
    """
    _check_group_exists(group_id)

    group = _server.Group(group_id=group_id, description=description,
                          status=_server.Group.INACTIVE)
    _server.Group.add(group)
    _LOGGER.debug("Added group (%s).", group)

@_events.on_event(ACTIVATE_GROUP)
def _activate_group(group_id):
    """Activate a group.
    """
    group = _retrieve_group(group_id)
    group.status = _server.Group.ACTIVE

    _detector.FailureDetector.register_group(group_id)
    _LOGGER.debug("Group (%s) is active.", group)

@_events.on_event(DEACTIVATE_GROUP)
def _deactivate_group(group_id):
    """Deactivate a group.
    """
    group = _retrieve_group(group_id)
    group.status = _server.Group.INACTIVE
    _detector.FailureDetector.unregister_group(group_id)
    _LOGGER.debug("Group (%s) is active.", str(group))

@_events.on_event(UPDATE_GROUP)
def _update_group_description(group_id, description):
    """Update a group description.
    """
    group = _retrieve_group(group_id)
    group.description = description
    _LOGGER.debug("Updated group (%s).", group)

@_events.on_event(DESTROY_GROUP)
def _destroy_group(group_id, force):
    """Destroy a group.
    """
    group = _retrieve_group(group_id)

    _check_shard_exists(group_id)

    servers = group.servers()
    if servers and force:
        for server in servers:
            _server.MySQLServer.remove(server)
    elif servers:
        raise _errors.GroupError("Group (%s) is not empty." % (group_id, ))

    _detector.FailureDetector.unregister_group(group_id)
    _server.Group.remove(group)
    _LOGGER.debug("Removed group (%s).", group)

def _lookup_servers(group_id, server_id=None, status=None, mode=None):
    """Return existing servers in a group or information on a server.
    """
    group = _retrieve_group(group_id)

    status = _retrieve_server_status(status) if status is not None else None
    if status is None:
        status = _server.MySQLServer.SERVER_STATUS
    else:
        status = [status]

    mode = _retrieve_server_mode(mode) if mode is not None else None
    if mode is None:
        mode = _server.MySQLServer.SERVER_MODE
    else:
        mode = [mode]

    info = []
    if server_id is None:
        for server in group.servers():
            if server.status in status and server.mode in mode:
                _server_information(server, info)
    else:
        server = _retrieve_server(server_id, group_id)
        _server_information(server, info)

    return info

def _server_information(server, info):
    """Get information on server and append it into to a list.
    """
    info.append({
        "server_uuid" : str(server.uuid),
        "address" : server.address,
        "status" : server.status,
        "mode" : server.mode,
        "weight" : server.weight,
    })

def _lookup_uuid(address, timeout):
    """Return server's uuid.
    """
    try:
        return _server.MySQLServer.discover_uuid(
            address=address, connection_timeout=timeout
        )
    except _errors.DatabaseError:
        raise _errors.ServerError("Error accessing server (%s)." % (address, ))

@_events.on_event(ADD_SERVER)
def _add_server(group_id, address, timeout, update_only):
    """Add a server into a group.
    """
    group = _retrieve_group(group_id)
    uuid = _lookup_uuid(address, timeout)
    _check_server_exists(uuid)
    server = _server.MySQLServer(uuid=_uuid.UUID(uuid), address=address)

    # Check if the server fulfils the necessary requirements to become
    # a member.
    _check_requirements(server)

    # Add server to the state store.
    _server.MySQLServer.add(server)

    # Add server as a member in the group.
    server.group_id = group_id

    if not update_only:
        # Configure the server as a slave if there is a master.
        _configure_as_slave(group, server)

    _LOGGER.debug("Added server (%s) to group (%s).", server, group)

@_events.on_event(REMOVE_SERVER)
def _remove_server(group_id, server_id):
    """Remove a server from a group.
    """
    group = _retrieve_group(group_id)
    server = _retrieve_server(server_id, group_id)

    if group.master == server.uuid:
        raise _errors.ServerError(
            "Cannot remove server (%s), which is master in group (%s). "
            "Please, demote it first." % (server.uuid, group_id)
        )

    _server.MySQLServer.remove(server)
    server.disconnect()
    _server.ConnectionPool().purge_connections(server.uuid)

@_events.on_event(SET_SERVER_STATUS)
def _set_server_status(server_id, status, update_only):
    """Set a server's status.
    """
    status = _retrieve_server_status(status)
    server = _retrieve_server(server_id)

    if status == _server.MySQLServer.PRIMARY:
        _set_server_status_primary(server, update_only)
    elif status == _server.MySQLServer.FAULTY:
        _set_server_status_faulty(server, update_only)
    elif status == _server.MySQLServer.SECONDARY:
        _set_server_status_secondary(server, update_only)
    elif status == _server.MySQLServer.SPARE:
        _set_server_status_spare(server, update_only)

def _retrieve_server_status(status):
    """Check whether the server's status is valid or not and
    if an integer was provided retrieve the correspondent
    string.
    """
    valid = False
    try:
        idx = int(status)
        try:
            status = _server.MySQLServer.get_status(idx)
            valid = True
        except IndexError:
            pass
    except ValueError:
        try:
            status = str(status).upper()
            _server.MySQLServer.get_status_idx(status)
            valid = True
        except ValueError:
            pass

    if not valid:
        values = [ str((_server.MySQLServer.get_status_idx(value), value))
                   for value in _server.MySQLServer.SERVER_STATUS ]
        raise _errors.ServerError("Trying to use an invalid status (%s). "
            "Possible values are %s." % (status, ", ".join(values))
        )

    return status

def _set_server_status_primary(server, update_only):
    """Set server's status to primary.
    """
    raise _errors.ServerError(
        "If you want to make a server (%s) primary, please, use the "
        "group.promote function." % (server.uuid, )
    )

def _set_server_status_faulty(server, update_only):
    raise _errors.ServerError(
        "If you want to set a server (%s) to faulty, please, use the "
        "threat.report_faulty interface." % (server.uuid, )
    )

def _set_server_status_secondary(server, update_only):
    """Set server's status to secondary.
    """
    allowed_status = [_server.MySQLServer.SPARE]
    status = _server.MySQLServer.SECONDARY
    mode = _server.MySQLServer.READ_ONLY
    _do_set_status(server, allowed_status, status, mode, update_only)

def _set_server_status_spare(server, update_only):
    """Set server's status to spare.
    """
    allowed_status = [
        _server.MySQLServer.SECONDARY, _server.MySQLServer.FAULTY
    ]
    status = _server.MySQLServer.SPARE
    mode = _server.MySQLServer.OFFLINE
    previous_status = server.status
    _do_set_status(server, allowed_status, status, mode, update_only)

    if previous_status == _server.MySQLServer.FAULTY:
        # Check whether the server is really alive or not.
        _check_requirements(server)

        # Configure replication
        if not update_only:
             group = _server.Group.fetch(server.group_id)
             _configure_as_slave(group, server)

def _do_set_status(server, allowed_status, status, mode, update_only):
    """Set server's status.
    """
    allowed_transition = server.status in allowed_status
    previous_status = server.status
    if allowed_transition:
        server.status = status
        server.mode = mode
    else:
        raise _errors.ServerError(
            "Cannot change server's (%s) status from (%s) to (%s)." %
            (str(server.uuid), server.status, status)
        )

    _LOGGER.debug(
        "Changed server's status (%s) from (%s) to (%s). Update-only "
        "state store parameter is (%s).", str(server.uuid), previous_status,
        server.status, update_only
    )

@_events.on_event(SET_SERVER_WEIGHT)
def _set_server_weight(server_id, weight):
    """Set server's weight.
    """
    server = _retrieve_server(server_id)

    try:
        weight = float(weight)
    except ValueError:
        raise _errors.ServerError("Value (%s) must be a float." % (weight, ))

    if weight <= 0.0:
        raise _errors.ServerError(
            "Cannot set the server's weight (%s) to a value lower "
            "than or equal to 0.0" % (weight, )
        )
    server.weight = weight

@_events.on_event(SET_SERVER_MODE)
def _set_server_mode(server_id, mode):
    """Set server's mode.
    """
    mode = _retrieve_server_mode(mode)
    server = _retrieve_server(server_id)

    if server.status == _server.MySQLServer.PRIMARY:
        _set_server_mode_primary(server, mode)
    elif server.status == _server.MySQLServer.SECONDARY:
        _set_server_mode_secondary(server, mode)
    elif server.status == _server.MySQLServer.SPARE:
        _set_server_mode_spare(server, mode)
    elif server.status == _server.MySQLServer.FAULTY:
        _set_server_mode_faulty(server, mode)

@_events.on_event(BACKUP_SERVER)
def _backup_server(source_uuid, host, port, mysqldump_binary,
                   mysqlclient_binary, config_file):
    """Backup the source server, given by the source_uuid.

    :param source_uuid: The UUID of the source server.
    :param host: The hostname of the destination server.
    :param port: The port number of the destination server.
    :param mysqldump_binary: The MySQL Dump Binary path.
    :param mysqlclient_binary: The MySQL Client Binary path.
    :param config_file: The complete path to the fabric configuration
        file.
    """
    source_server = _server.MySQLServer.fetch(source_uuid)
    #Do the backup of the group hosting the source shard.
    backup_image = _backup.MySQLDump.backup(
                        source_server,
                        config_file,
                        mysqldump_binary
                    )
    _LOGGER.debug("Done with backup of server with uuid = %s.", source_uuid)
    procedures = _events.trigger_within_procedure(
        RESTORE_SERVER,
        source_uuid,
        host,
        port,
        backup_image.path,
        mysqlclient_binary,
        config_file
    )

@_events.on_event(RESTORE_SERVER)
def _restore_server(source_uuid, host, port, backup_image, mysqlclient_binary,
                    config_file):
    """Restore the backup on the destination Server.

    :param source_uuid: The UUID of the source server.
    :param host: The hostname of the destination server.
    :param port: The port number of the destination server.
    :param userid: The User Name to be used to connect to the destn server.
    :param passwd: The password to be used to connect to the destn server.
    :param mysqldump_binary: The MySQL Dump Binary path.
    :param mysqlclient_binary: The MySQL Client Binary path.
    :param config_file: The complete path to the fabric configuration
        file.
    """
    source_server = _server.MySQLServer.fetch(source_uuid)
    if not source_server:
        raise _errors.ServerError(SERVER_NOT_FOUND % source_uuid)
    #Build a backup image that will be used for restoring
    bk_img = _backup.BackupImage(backup_image)
    _backup.MySQLDump.restore_server(
        host,
        port,
        source_server.USER,
        source_server.PASSWD,
        bk_img,
        config_file,
        mysqlclient_binary
    )
    _LOGGER.debug("Done with restore of server with host = %s, port = %s" %\
                  (host, port,))

def _retrieve_server_mode(mode):
    """Check whether the server's mode is valid or not and if an integer was
    provided retrieve the correspondent string.
    """
    valid = False
    try:
        idx = int(mode)
        try:
            mode = _server.MySQLServer.get_mode(idx)
            valid = True
        except IndexError:
            pass
    except ValueError:
        try:
            mode = str(mode).upper()
            _server.MySQLServer.get_mode_idx(mode)
            valid = True
        except ValueError:
            pass

    if not valid:
        values = [ str((_server.MySQLServer.get_mode_idx(value), value))
                   for value in _server.MySQLServer.SERVER_MODE ]
        raise _errors.ServerError("Trying to use an invalid mode (%s). "
            "Possible values are: %s." % (mode, ", ".join(values))
        )

    return mode

def _set_server_mode_primary(server, mode):
    """Set server's mode when it is a primary.
    """
    allowed_mode = \
        (_server.MySQLServer.WRITE_ONLY, _server.MySQLServer.READ_WRITE)
    _do_set_server_mode(server, mode, allowed_mode)

def _set_server_mode_secondary(server, mode):
    """Set server's mode when it is a secondary.
    """
    allowed_mode = \
        (_server.MySQLServer.OFFLINE, _server.MySQLServer.READ_ONLY)
    _do_set_server_mode(server, mode, allowed_mode)

def _set_server_mode_spare(server, mode):
    """Set server's mode when it is a spare.
    """
    allowed_mode = \
        (_server.MySQLServer.OFFLINE, _server.MySQLServer.READ_ONLY)
    _do_set_server_mode(server, mode, allowed_mode)

def _set_server_mode_faulty(server, mode):
    """Set server's mode when it is a faulty.
    """
    allowed_mode = ()
    _do_set_server_mode(server, mode, allowed_mode)

def _do_set_server_mode(server, mode, allowed_mode):
    """Set server's mode.
    """
    if mode not in allowed_mode:
        raise _errors.ServerError(
            "Cannot set mode to (%s) when the server's (%s) status is (%s)."
            % (mode, server.uuid, server.status)
        )
    server.mode = mode

def _retrieve_server(server_id, group_id=None):
    """Return a MySQLServer object from a UUID or an HOST:PORT.
    """
    uuid = _retrieve_uuid_object(server_id)

    server = _server.MySQLServer.fetch(uuid)
    if not server:
        raise _errors.ServerError(
            "Server (%s) does not exist." % (uuid, )
            )

    if not server.group_id:
        raise _errors.GroupError(
            "Server (%s) does not belong to a group." % (uuid, )
            )

    if group_id is not None and group_id != server.group_id:
        raise _errors.GroupError(
            "Group (%s) does not contain server (%s)." % (group_id, uuid)
        )
    return server

def _check_server_exists(server_id):
    """Check whether a MySQLServer instance exists or not.

    :param server_id: The UUID or the host:port for the server.
    """
    server_id = _retrieve_uuid_object(server_id)

    server = _server.MySQLServer.fetch(server_id)
    if server:
        raise _errors.ServerError("Server (%s) already exists." % (server_id, ))

def _retrieve_group(group_id):
    """Return a Group object from an identifier.
    """
    group = _server.Group.fetch(group_id)
    if not group:
        raise _errors.GroupError("Group (%s) does not exist." % (group_id, ))
    return group

def _check_group_exists(group_id):
    """Check whether a group exists or not.
    """
    group = _server.Group.fetch(group_id)
    if group:
        raise _errors.GroupError("Group (%s) already exists." % (group_id, ))

def _check_shard_exists(group_id):
    """Check whether there is a shard associated with the group.
    """
    shard_id = _sharding.Shards.lookup_shard_id(group_id)
    if shard_id:
        raise _errors.GroupError(
            "Cannot erase a group (%s) which is associated to a shard (%s)." % 
            (group_id, shard_id)
        )

    shard_mapping_id = _sharding.ShardMapping.lookup_shard_mapping_id(group_id)
    if shard_mapping_id:
        raise _errors.GroupError(
            "Cannot erase a group (%s) which is used as a global group in a "
            "shard definition (%s)." % (group_id, shard_mapping_id)
        )

def _retrieve_uuid_object(server_id):
    """Transform an input string into a UUID object.
    """
    assert(isinstance(server_id, basestring))

    try:
        return _uuid.UUID(server_id)
    except ValueError:
        pass

    try:
        return _server.MySQLServer.discover_uuid(address=server_id)
    except ValueError:
        raise _errors.ServerError(
            "Error trying to access server identified by (%s)." % (server_id, )
        )

def _check_requirements(server):
    """Check if the server fulfils some requirements.
    """
    # Being able to connect to the server is the first requirment.
    server.connect()

    if not server.check_version_compat((5, 6, 8)):
        raise _errors.ServerError(
            "Server (%s) has an outdated version (%s). 5.6.8 or greater "
            "is required." % (server.uuid, server.version)
            )

    if not server.has_required_privileges():
        raise _errors.ServerError(
            "User (%s) does not have appropriate privileges (%s) on server "
            "(%s, %s)." % (server.user,
            ", ".join(_server.MySQLServer.ALL_PRIVILEGES),
            server.address, server.uuid)
        )

    if not server.gtid_enabled or not server.binlog_enabled:
        raise _errors.ServerError(
            "Server (%s) does not have the binary log or gtid enabled."
            % (server.uuid, )
            )

def _configure_as_slave(group, server):
    """Configure the server as a slave.
    """
    try:
        if group.master:
            master = _server.MySQLServer.fetch(group.master)
            master.connect()
            _utils.switch_master(server, master)
    except _errors.DatabaseError as error:
        _LOGGER.debug(
            "Error configuring slave (%s)...", server.uuid, exc_info=error
        )
        raise _errors.ServerError(
            "Error trying to configure Server (%s) as slave."
            % (server.uuid, )
        )
